# coding: utf-8

# ref: http://www.piece-me.org/piece-lab/adpcm/adpcm4.html
#      P/ECE ADPCM

require "stringio"

module IOExt
	def readw; read(2).unpack("v")[0]; end
	def readd; read(4).unpack("V")[0]; end

	def writew v; write [v].pack("v"); end
	def writed v; write [v].pack("V"); end
end

class ADPCMS16PIECE
	attr_accessor :val, :delta

	SCALE = [233, 233, 233, 233, 310, 375, 453, 549] # /256

	def initialize
		@presample = 0
		@predict = 16
	end

	def next_scale dscale
		@predict * SCALE[dscale.abs] / 256
	end

	def encode sample
		delta = sample - @presample
		dscale = [[(delta.to_f / @predict).round, 7].min, -7].max
		if next_scale(dscale) < 16
			dscale = (0 <= dscale) ? -4 : 4 # +-4 = reversed least bigger than 256
		end

		# update state by decoding myself
		decode_dscale dscale

		dscale + 8
	end

	def decode sample
		decode_dscale sample - 8
	end

	def self.encode samples
		encoder = self.new
		samples.map { |e| encoder.encode e }
	end
	def self.decode samples
		decoder = self.new
		samples.map { |e| decoder.decode e }
	end

	private

	def decode_dscale dscale
		@presample = @presample + @predict * dscale
		@predict = next_scale dscale
		@presample
	end
end

class WAVPCMS16MONO
	attr_accessor :type, :channels, :srate, :avgbps, :blksize, :bitpersample, :samples

	def initialize
		@type = 1 # PCM
		@channels = 1 # Mono
		@srate = nil
		@avgbps = nil
		@blksize = nil
		@bitpersample = 16 # 16bit(le)
		@samples = []
	end

	def read io
		io.extend IOExt

		io.read(4) == "RIFF" or raise "?RIFF signature not found"
		io.readd
		io.read(4) == "WAVE" or raise "?type is not WAVE"

		io.read(4) == "fmt " or raise "?no fmt chunk"
		io.readd == 16 or raise "?fmt chunk length != 16"
		type = io.readw
		type == 1 or raise "?format_id is not 1 (PCM)"
		channels = io.readw
		channels == 1 or raise "?not monoral"
		srate = io.readd
		avgbps = io.readd
		blksize = io.readw
		bitpersample = io.readw
		bitpersample == 16 or raise "?byte-per-sample is not 16"

		io.read(4) == "data" or raise "?no following data chunk"
		datalen = io.readd
		data = io.read(datalen)

		# ignore trailing bytes

		# set fields
		@type = type
		@channels = channels
		@srate = srate
		@avgbps = avgbps
		@blksize = blksize
		@bitpersample = bitpersample
		@samples = data.unpack("s*") # s16(le) but no such format in String#unpack, use s16(native-endian)

		self
	end

	def autoheader
		@avgbps ||= @srate * @channels * @bitpersample / 8
		@blksize ||= @channels * @bitpersample / 8
	end

	def write io
		autoheader

		data = @samples.pack("s*")
		ck_data = ["data", data.size, data].pack("a*Va*")

		fmt = StringIO.open("".b) { |io|
			io.extend IOExt
			io.writew @type
			io.writew @channels
			io.writed @srate
			io.writed @avgbps
			io.writew @blksize
			io.writew @bitpersample
			io.string # return its string before auto-closed
		}
		ck_fmt = ["fmt ", fmt.size, fmt].pack("a*Va*")

		all = "WAVE" + ck_fmt + ck_data

		io.extend IOExt
		io.write "RIFF"
		io.writed all.size
		io.write all

		self
	end
end

if $0 == __FILE__
	wav = open(ARGV[0], File::RDONLY | File::BINARY) { |f| WAVPCMS16MONO.new.read(f) }

	encoder = ADPCMS16PIECE.new
	decoder = ADPCMS16PIECE.new
	adpcmsamples = wav.samples.map { |e| encoder.encode e }

	open(ARGV[0].sub(/\.[^.]+$/, "") + "_adpcm.bin", File::WRONLY | File::CREAT | File::BINARY) { |f|
		f.write [adpcmsamples.map{|e| "%X" % e}.join + "0"].pack("h*")
	}

	pcmsamples = adpcmsamples.map { |e| decoder.decode e }

	open(ARGV[0].sub(/\.[^.]+$/, "") + "_adpcm.wav", File::WRONLY | File::CREAT |
	#File::EXCL |
	File::BINARY) { |f|
		w = WAVPCMS16MONO.new
		w.srate = wav.srate
		w.samples = pcmsamples
		w.write f
	}

	open(ARGV[0].sub(/\.[^.]+$/, "") + "_adpcm_resample.wav", File::WRONLY | File::CREAT |
	#File::EXCL |
	File::BINARY) { |f|
		w = WAVPCMS16MONO.new
		w.srate = 44100
		w.samples = (pcmsamples.size * w.srate / wav.srate).times.map { |i| pcmsamples[i * wav.srate / w.srate] }
		w.write f
	}
end
