	.section .text.start, "ax", %progbits

	.globl start
	.type start, %function
start:
	// copy initialized data
	ldr r0, =__data_start
	ldr r1, =__data_end
	ldr r2, =__data_from
1:
	cmp r0, r1
#ifdef __THUMBEL__
	bge 1f
	ldr r3, [r2]
	str r2, [r0]
	add r2, #4
	add r0, #4
	b 1b
1:
#elif defined(__arm__)
	ldrlt r3, [r2], #4
	strlt r2, [r0], #4
	blt 1b
#else
#error what? neither ARM or THUMB??
#endif

	// initialize bss: kernel may already cleaned, but do this...
	ldr r0, =__bss_start
	ldr r1, =__bss_end
	mov r2, #0
1:
	cmp r0, r1
#ifdef __THUMBEL__
	bge 1f
	str r2, [r0]
	add r0, #4
	b 1b
1:
#elif defined(__arm__)
	strlt r2, [r0], #4
	blt 1b
#else
#error what? neither ARM or THUMB??
#endif

	// register handlers...
	mov r0, #0; ldr r1, =c_gate_swi; svc 0x01
	mov r0, #1; ldr r1, =c_gate_irq; svc 0x01
	mov r0, #2; ldr r1, =c_gate_fiq; svc 0x01
	mov r0, #3; ldr r1, =c_gate_download; svc 0x01

	// call main
	ldr r0, =appmain
#ifdef __THUMBEL__
	mov r1, pc
	add r1, #5
	mov lr, r1
#elif defined(__arm__)
	mov lr, pc
#else
#error what? neither ARM or THUMB??
#endif
	bx r0

	// return to menu program with our program selected
	svc 0x16 //getcurrent

#ifdef __THUMBEL__
	mov r2, #0x30 // param: 0x30 | dir_index = show with selected that
	orr r2, r0
#elif defined(__arm__)
	orr r2, r0, #0x30
#else
#error what? neither ARM or THUMB??
#endif
	mov r0, #1 // flag
	mov r1, #0 // dir_index: menu
	svc 0x08 //prepareexecute

	mov r0, #0 // no snapshot
	svc 0x09 //doexecute
	// noreturn... if not misprogrammed to reach here with COM enabled.
	b . // safe-net

.macro make_gate suffix, offset
	.section .text.c_gate_\suffix, "ax", %progbits
	.globl c_gate_\suffix
	.type c_gate_\suffix, %function
c_gate_\suffix:
	push {r0, r1}
	mov r0, #\offset
	b c_gate_bottom_half
.endm

make_gate swi, 0
make_gate irq, 4
make_gate fiq, 8
make_gate download, 12

	.section .text.c_gate_bottom_half, "ax", %progbits
	.type c_gate_bottom_half, %function
c_gate_bottom_half:
	ldr r1, =low_handler_ptrs
	add r1, r0
	ldr r0, [r1]
	tst r0, r0
	beq 1f

#ifdef __THUMBEL__
	push {r2, r3}
	mov r2, r12
	push {r2}
	push {lr}
	mov r1, pc
	add r1, #5
	mov lr, r1
#elif defined(__arm__)
	push {r2, r3, r12, lr}
	mov lr, pc
#else
#error what? neither ARM or THUMB??
#endif
	bx r0
#ifdef __THUMBEL__
	pop {r2}
	mov lr, r2
	pop {r2}
	mov r12, r2
	pop {r2, r3}
#elif defined(__arm__)
	pop {r2, r3, r12, lr}
#else
#error what? neither ARM or THUMB??
#endif

1:
	pop {r0, r1}
	bx lr

	.section .bss.low_handler_ptrs, ""
	.balign 4
	.globl low_handler_ptrs
low_handler_ptrs: .space 4 * 4
