#include "../krom_c.h"

#define SYSSETCB_SWI 0
#define SYSSETCB_IRQ 1
#define SYSSETCB_FIQ 2
#define SYSSETCB_CUSTOM3 3

typedef void *Scene(u32 stat, u32 prevstat);

struct _range {
	u32 min, max;
};

// gcc may assumes volatile if no output, but explicitly add volatile to __asm__.
static inline void sysSetCallbacks(u32 index, void *pfunc) { __asm__ volatile("mov r0, %0; mov r1, %1; svc 0x01" : : "rI"(index), "rI"(pfunc) : "r0", "r1"); }
// gcc may assumes volatile if no output, but explicitly add volatile to __asm__.
static inline void sysSetCpuSpeed(u32 speed) { __asm__ volatile("mov r0, %0; svc 0x04" : : "rI"(speed) : "r0"); }
// gcc may assumes volatile if no output, but explicitly add volatile to __asm__.
static inline void sysSenseAutoCom() { __asm__ volatile("svc 0x05" : : : "r0"); }
// note: return value of GetPtrToKernFlags(ComFlags) is run-time constant. adding volatile to __asm__ makes code redundant.
// TODO how to return r0? is this best way?
static inline u32 *sysGetPtrToComFlags() { void *result; __asm__("svc 0x06; mov %0, r0" : "=r"(result) : : "r0"); return result; }
// calling PrepareExecute changes kernel state, so this cannot be optimized. specify volatile to __asm__.
static inline u32 sysPrepareExecute(u32 doSet, u16 appFirstBlock, u32 param) { u32 result; __asm__ volatile("mov r0, %1; mov r1, %2; mov r2, %3; svc 0x08; mov %0, r0" : "=r"(result) : "rI"(doSet), "rI"(appFirstBlock), "rI"(param) : "r0", "r1", "r2", "r3"); return result; }
// this is almost noreturn. returns if COM enabled.
// gcc may assumes volatile if no output, but explicitly add volatile to __asm__.
static inline void sysDoExecute(u32 doSnapshotSave) { __asm__ volatile("mov r0, %0; svc 0x09" : : "rI"(doSnapshotSave) : "r0"); }
// gcc may assumes volatile if no output, but explicitly add volatile to __asm__.
static inline void sysSetBcdDateTime(u32 date, u32 time) { __asm__ volatile("mov r0, %0; mov r1, %1; svc 0x0c" : : "rI"(date), "rI"(time) : "r0", "r1"); }
// return value will differ between calls, specify volatile to __asm__.
static inline u32 sysGetBcdDate() { u32 result; __asm__ volatile("svc 0x0d; mov %0, r0" : "=r"(result) : : "r0"); return result; }
// return value will differ between calls, specify volatile to __asm__.
static inline u32 sysGetBcdTime() { u32 result; __asm__ volatile("svc 0x0e; mov %0, r0" : "=r"(result) : : "r0"); return result; }
// gcc may assumes volatile if no output, but explicitly add volatile to __asm__.
static inline void sysSetComOnOff(u32 on) { __asm__ volatile("mov r0, %0; svc 0x11" : : "rI"(on) : "r0"); }
// note: return value of GetPtrToAlarmSetting is run-time constant. adding volatile to __asm__ makes code redundant.
// TODO how to return r0? is this best way?
static inline struct alarm_info *sysGetPtrToAlarmSetting() { void *result; __asm__("svc 0x13; mov %0, r0" : "=r"(result) : : "r0"); return result; }
