#include "common_c.h"

void irqHandler();
void fiqHandler();
void custom3Handler();
void *scene_clock_entry(u32 stat, u32 prevstat);
void *scene_app_entry(u32 stat, u32 prevstat);

volatile int VSync;
u8 NumUsingBlocks;
u8 NumFiles;
u8 NumApps;
u8 FileIdx2FirstBlock[15]; // for fileviewer
u8 AppIdx2FirstBlock[15]; // for appviewer
u32 OffScreen[32];
u32 *PComFlags;
struct alarm_info *PAlarmInfo;
volatile u32 sndlen2;
s8 GuiParamApp;

const u32 font4x6[] = { // LSByte most top, LSbit most left
	0x00025552, // 0
	0x00022232, // 1
	0x00071243, // 2
	0x00034243, // 3
	0x00044755, // 4
	0x00034317, // 5
	0x00025316, // 6
	0x00022247, // 7
	0x00025252, // 8
	0x00034652, // 9

	0x00034216, // 10 S
	0x00063520, // 11 e
	0x00042720, // 12 t
	0x00022202, // 13 i
	0x00055300, // 14 n
	0x00036530, // 15 g
	0x00061160, // 16 c
	0x00042222, // 17 l
	0x00025520, // 18 o
	0x00053511, // 19 k
	0x00050000, // 20 ..

	0x00065500, // 21 u
	0x00055775, // 22 M
	0x00022227, // 23 T
	0x00057755, // 24 W
	0x00055311, // 25 h
	0x00011717, // 26 F
	0x00013500, // 27 r
	0x00065600, // 28 a

	0x00875578, // 29 spk
	0x00925978, // 30 spk off left
	0x00000012, // 31 spk off right
	0x00001100, // 32 wave1
	0x00045540, // 33 wave2
	0x00844488, // 34 alarm on left
	0x00347532, // 35 alarm on right
	0x00484488, // 36 alarm off left
	0x00345252, // 37 alarm off right

	0x00012244, // 38 slash
};

// TODO copied from kernel/lib_c.c, dedup is better.
static inline u32 bcd2int2(u32 bcd) {
	return ((bcd >> 4) * 10) + (bcd & 0x0F);
}

void handleIrq() {
	u32 intr = *I_HOLD & *I_MASK;
	if(intr & INTMASK_TIMER0) {
		VSync = 1;
		*I_HCLR = INTMASK_TIMER0;
	}
	if(intr & INTMASK_TIMER1) {
		if(sndlen2 == 0) {
			TIMER[1].mode = TIMER_DISABLE;
			*DAC_CTRL = DACCTRL_DISABLE;
			*IOP_STOP = IOPPIN_SOUND;
		} else {
			sndlen2--;
			s16 data;
			if(sndlen2 % 2 == 0) {
				data = -0x8000;
			} else {
				data = 0x7FFF;
			}
			*DAC_DATA = data >> (PAlarmInfo->attenate * 2);
		}
		*I_HCLR = INTMASK_TIMER1;
	}
	if(intr & INTMASK_RTC) {
		*I_HCLR = INTMASK_RTC;
	}
	if((intr & INTMASK_IOP) || (intr & INTMASK_ACTION)) {
		// TODO reset idle time
		//*I_HCLR = INTMASK_IOP; // DO NOT CLEAR HOLD IOP HERE! kernel handles!
	}
	if(intr & INTMASK_BATTLO) {
		// TODO clock down, stop sound.
		*I_HCLR = INTMASK_BATTLO;
	}
}

void handleFiq() {
}

void handleCustom3() {
}

static inline void setTimer(int timer, u32 mode, u32 max) {
	TIMER[timer].mode = TIMER_DISABLE; // or setting max will be ignored
	TIMER[timer].max = max;
	TIMER[timer].mode = TIMER_ENABLE | mode;
}

// if sec is constant, libgcc is not required.
static inline void setTimerBySec(int timer, u32 clkmode, float sec, int powersaving) {
	u32 cpuclkbase = (clkmode < 6) ? 0x7C00 : 0x7A00;
	u32 cpuclkhz = cpuclkbase << clkmode;
	u32 mode;
	u32 max;
	if(!powersaving) {
		// best effort for precision
		if((u32)(sec * cpuclkhz / 2) <= 0xFFFF) {
			mode = TIMER_DIV2;
			max = (u32)(sec * cpuclkhz / 2);
		} else if((u32)(sec * cpuclkhz / 32) <= 0xFFFF) {
			mode = TIMER_DIV32;
			max = (u32)(sec * cpuclkhz / 32);
		} else if((u32)(sec * cpuclkhz / 512) <= 0xFFFF) {
			mode = TIMER_DIV512;
			max = (u32)(sec * cpuclkhz / 512);
		} else {
			// too long... (>~8.39s in clkmode7)
			mode = TIMER_DIV512;
			max = 0xFFFF;
		}
	} else {
		// powersaving: best effort for high-div
		if(0xFFFF < (u32)(sec * cpuclkhz / 512)) {
			// too long... (>~8.39s in clkmode7)
			mode = TIMER_DIV512;
			max = 0xFFFF;
		} else if(0 < (u32)(sec * cpuclkhz / 512)) {
			mode = TIMER_DIV512;
			max = (u32)(sec * cpuclkhz / 512);
		} else if(0 < (u32)(sec * cpuclkhz / 32)) {
			mode = TIMER_DIV32;
			max = (u32)(sec * cpuclkhz / 32);
		} else {
			// too short for div512/div32...
			mode = TIMER_DIV2;
			max = (u32)(sec * cpuclkhz / 2);
		}
	}
	setTimer(timer, mode, max);
}

void readMCDirectory() {
	NumUsingBlocks = 0;
	NumFiles = 0;
	NumApps = 0;

	if(*((volatile u16*)((u8*)PFLA + 0)) != 0x434D/*MC*/) {
		// no valid memorycard magic exists
		return;
	}

	// stage 1: scan block 1..15 for file existence
	for(u32 i = 1; i <= 15; i++) {
		const struct mcdir *ent = (const struct mcdir *)((u8*)PFLA + i * 0x80);
		if((ent->state & 0xF0) == 0x50) { // some using block
			NumUsingBlocks++;
			if(ent->state == 0x51) { // first using block of some file
				FileIdx2FirstBlock[NumFiles] = i;
				NumFiles++;
			}
		}
	}

	// stage 2: scan first blocks of files for valid(-looking) app
	for(u32 i = 0; i < NumFiles; i++) {
		const struct mcdir *ent = (const struct mcdir *)((u8*)PFLA + FileIdx2FirstBlock[i] * 0x80);
		if(ent->name[6] != 'P') {
			// this file is not for pk.
			continue;
		}
		const struct schdr *hdr = (const struct schdr *)((u8*)PFLA + FileIdx2FirstBlock[i] * 0x2000);
		if((hdr->pkmagic & 0xFEFFffff) != 0x3058434D/*MCX0*/) {
			// this file does not contain valid (normal or snapshot) pk magic.
			continue;
		}

		// looks valid; add this file to app list.
		AppIdx2FirstBlock[NumApps] = FileIdx2FirstBlock[i];
		NumApps++;
	}
}

// @return week (1=Sunday, 2=Monday, ..., 7=Saturday; same as RTC week)
// note: century and year2 are separated because dividing integer requires libgcc that does not support APCS...
u32 calcWeekBCD(const u32 centuryBCD, const u32 year2BCD, const u32 monthBCD, const u32 dayBCD) {
	/***/ u32 century = bcd2int2(centuryBCD);
	const u32 year2 = bcd2int2(year2BCD);
	const u32 month = bcd2int2(monthBCD);
	const u32 day = bcd2int2(dayBCD);

	//static const u32 mparams[12] = {98, 101, 100, 61, 112, 101, 110, 43, 109, 97, 100, 46};
	static const u8 mparams[12] = {0, 3, 2, 5, 0, 3, 5, 1, 4, 6, 2, 4};

	const u32 yprime = (century * 100 + year2) - (month < 3);
	if(year2 == 0) { // year % 100 == 0... TODO: when 400?
		century -= month < 3;
	}
	// note: 365%7==1
	u32 dprime = yprime / 4/*leap4*/ + yprime/*year*/ - century/*leap100*/ + century / 4/*leap400*/ + mparams[month - 1] + day;
#if 0
	// requires libgcc
	return 1 + (dprime % 7);
#elif 0
	// original: safe but slow
	while(7 <= dprime) {
		dprime -= 7;
	}
	return 1 + dprime;
#elif 1
	// manually doing u32.32 fixed-point divide
	// no volatile, let gcc "optimize" for (or "make runnable" on) PK201.
	const u32 r = (u32)((1ULL << 32) / 7.) + 1; // +1 = ceil
	const u64 tmp = (u64)r * dprime;
	return 1 + (dprime - (tmp >> 32) * 7);
#elif 0
	// manually doing u32.32 fixed-point divide
	// note: use volatile or gcc don't use umull (that I think faster and smaller than many add/lsl/...)...?
	//       but redundant memory access is also generated...
	// note2: PK201 does not implement umull/smull. dst regs are totally UNchanged.
	volatile u32 r = (1ULL << 32) / 7.;
	volatile u64 tmp = (u64)r * dprime;
	return 1 + (dprime - (tmp >> 32) * 7);
#elif 0
	__asm__ volatile("ldr r0, =1234567890; mov r1, #34; mov r2, #56; mov r3, #0x11000000; mov r4, #0x22000000; mov r5, #0x33000000; mov r6, #0x44000000; mov r7, #0x55000000; mla r3, r0, r1, r2; umull r4, r5, r0, r1; smull r6, r7, r0, r1" : : : "r0", "r1", "r2", "r3", "r4", "r5", "r6", "r7");
	for(;;);
#endif
}

void setRTC(const u32 datebcd, const u32 timebcd) {
	// sysSetBcdDateTime takes some time. display "Setting clock...."
	// clearing top 12 lines
	//for(u32 i = 0; i < 12; i++) { VRAM[i] = 0x00000000; }
	for(u32 y = 0; y < 6; y++) {
		{
			u32 vl;// = VRAM[0+y];
			vl  = ((font4x6[10] >> (y * 4)) & 0x0F) <<  0;
			vl |= ((font4x6[11] >> (y * 4)) & 0x0F) <<  4;
			vl |= ((font4x6[12] >> (y * 4)) & 0x0F) <<  8;
			vl |= ((font4x6[12] >> (y * 4)) & 0x0F) << 12;
			vl |= ((font4x6[13] >> (y * 4)) & 0x0F) << 16;
			vl |= ((font4x6[14] >> (y * 4)) & 0x0F) << 20;
			vl |= ((font4x6[15] >> (y * 4)) & 0x0F) << 24;
			VRAM[0+y] = vl;
		}
		{
			u32 vl;// = VRAM[6+y];
			vl  = ((font4x6[16] >> (y * 4)) & 0x0F) <<  0;
			vl |= ((font4x6[17] >> (y * 4)) & 0x0F) <<  4;
			vl |= ((font4x6[18] >> (y * 4)) & 0x0F) <<  8;
			vl |= ((font4x6[16] >> (y * 4)) & 0x0F) << 12;
			vl |= ((font4x6[19] >> (y * 4)) & 0x0F) << 16;
			vl |= ((font4x6[20] >> (y * 4)) & 0x0F) << 20;
			vl |= ((font4x6[20] >> (y * 4)) & 0x0F) << 24;
			VRAM[6+y] = vl;
		}
	}

	const u8 c = (datebcd >> 24) & 0xFF;
	const u8 y = (datebcd >> 16) & 0xFF;
	const u8 m = (datebcd >>  8) & 0xFF;
	const u8 d = (datebcd >>  0) & 0xFF;
	const u32 w = calcWeekBCD(c, y, m, d);
	sysSetBcdDateTime(datebcd, (timebcd & 0x00FFffff) | (w << 24));
}

void clearOffScreen() {
	for(u32 y = 0; y < 32; y++) {
		OffScreen[y] = 0x00000000;
	}
}

void flipScreen() {
	for(u32 y = 0; y < 32; y++) {
		VRAM[y] = OffScreen[y];
	}
}

void setCpuSpeed(u32 mode) {
	if((*CLK_MODE & 0x0F) == mode) {
		// already that speed. noop.
		return;
	}

	sysSetCpuSpeed(mode);
	setTimer(0, TIMER_DIV512, (u32)(3997696/*Hz in mode 7*/ / 512 * (1/30.0)) >> (7 - mode));
	for(VSync = 0; !VSync; ) { /*nothing*/ } // wait for stable timer0
}

void startBeep() {
	// TODO avoid race-condition about DAC_CTRL/IOP_START/setTimer.
	setCpuSpeed(CLKMODE_4MHZ);
	if(PAlarmInfo->attenate == 2) {
		// mute
		*DAC_CTRL = DACCTRL_DISABLE;
		*IOP_STOP = IOPPIN_SOUND;
		sndlen2 = 0;
	} else {
		*DAC_CTRL = DACCTRL_ENABLE;
		*IOP_START = IOPPIN_SOUND;
		const double hz = 440.0;
		sndlen2 = (u32)((hz * 2) * 0.02);
		setTimer(1, TIMER_DIV2, (u32)(/*clk7*/3997696 / /*div*/2 / (hz * 2)));
	}
}

void runApp(u32 appFirstBlock) {
	*F_CTRL = 0;
	TIMER[0].mode = TIMER_DISABLE;
	TIMER[1].mode = TIMER_DISABLE;
	TIMER[2].mode = TIMER_DISABLE; // unused but do
	*I_HCLR = INTMASK_TIMER0 | INTMASK_RTC | INTMASK_TIMER2;
	*DAC_CTRL = DACCTRL_DISABLE;
	*IOP_STOP = IOPPIN_SOUND;
	sysSetComOnOff(0);
	// TODO if app-param 0x100 set, skip setting alarm except fontptr.
	// appFirstBlock == 0 if prepared by BU
	if(appFirstBlock != 0) {
		sysPrepareExecute(1, appFirstBlock, 0);
	}
	sysDoExecute(0);
}

void guimain(u32 param) {
	*I_MCLR = ~(INTMASK_COM | INTMASK_IOP);

	sysSetCallbacks(SYSSETCB_IRQ, irqHandler);
	sysSetCallbacks(SYSSETCB_FIQ, fiqHandler);
	sysSetCallbacks(SYSSETCB_CUSTOM3, custom3Handler);

	//setCpuSpeed(CLKMODE_4MHZ);
	sysSetCpuSpeed(CLKMODE_4MHZ);
	setTimer(0, TIMER_DIV512, (u32)(3997696/*Hz in mode 7*/ / 512 * (1/30.0)) >> (7 - CLKMODE_4MHZ));
	TIMER[1].mode = TIMER_DISABLE;

	*IOP_STOP = IOPPIN_REDLED | IOPPIN_SOUND;
	*IOP_CTRL = 2; // only REDLED is output pin?
	*DAC_CTRL = DACCTRL_DISABLE;

	// keep rotating state but initialize others.
	*LCD_CTRL = (*LCD_CTRL & LCDCTRL_ROTATE) | LCDCTRL_ENABLE | LCDCTRL_ACTIVE | LCDCTRL_RR64HZ;

	PComFlags = sysGetPtrToComFlags();
	PAlarmInfo = sysGetPtrToAlarmSetting();

	*I_MASK = INTMASK_TIMER0 | INTMASK_TIMER1 | INTMASK_RTC | INTMASK_IOP;

	readMCDirectory();

	Scene *pScene = scene_clock_entry;

	if((param & 0xF0) == 0x10) {
		// speakerside
		GuiParamApp = 0;
	} else if(param == 0x20) {
		// clockside
		GuiParamApp = 0;
	} else if((param & 0xF0) == 0x20) {
		// adjust RTC and exec that app
		GuiParamApp = param & 0x0F;
	} else if((param & 0xF0) == 0x30) {
		// select that app
		pScene = scene_app_entry;
		GuiParamApp = param & 0x0F;
	} else {
		if(!PAlarmInfo->goodrtc) {
			setRTC(0x20160101, 0x000000);
			PAlarmInfo->goodrtc = 1;
			GuiParamApp = 0;
		} else {
			GuiParamApp = -1;
		}

		// special: when reset, if some apps are installed, auto-select first one.
		if(0 < NumApps) {
			pScene = scene_app_entry;
			GuiParamApp = 0;
		}
	}

	sysSetComOnOff(1);

	u32 prevstat = 0;

	for(;;) {
		sysSenseAutoCom();
		if(*PComFlags & KERNFLAGS_PREPARED) {
			// XXX original do sysSetComOnOff(0) here, but it is also done in runApp... timing?
			runApp(0);
		}
		u32 stat = *I_STAT;
		// TODO keylocking
		clearOffScreen();
		for(;;) {
			Scene *pNextScene = pScene(stat, prevstat);
			prevstat = stat; // to avoid switching clock->app->clock->...
			if(!pNextScene) {
				break;
			}
			pScene = pNextScene;
		}
		// TODO if keylock, some effect
		// TODO if MCRW, indicate it, should readMCDirectory
		flipScreen();

		// TODO sleep if idle

		for(VSync = 0; !VSync; ) { /*nothing*/ }
	}
}
