#include "common_c.h"

extern u8 NumApps;
extern u8 AppIdx2FirstBlock[15];
extern u32 OffScreen[32];
extern s8 GuiParamApp;

static u8 Selection;
static u8 IconIndex;
static u8 CurrentFrame;
static u8 AnimWait;
static struct exiconentry *ExIconEntries; // cache
static struct exiconentry ExIconFromFVIcon; // dynamically built for app without exicon.
static u32 PressedSinceSceneEnter;

extern const u32 font4x6[];

void startBeep();
void clearOffScreen();
void flipScreen();
void runApp(u32 appFirstBlock);
void *scene_clock_entry(u32 stat, u32 prevstat);

void mapApp(u32 appidx) {
	// unmap all
	*F_CTRL = 0; // not 1?
	*F_TAG = 0;

	/*
	for(u32 i = 0; i < 15; i++) {
		F_DATA[i] = 0x0F;
	}
	*/

	// then map app file
	u32 tag = 0;
	u32 block = AppIdx2FirstBlock[appidx];
	for(u32 i = 0; block < 16 && i < 15; i++) { // note: different condition than kernel's doExecute.
		struct mcdir *ent = (struct mcdir *)((u8*)PFLA + 0x80 * block);
		F_DATA[block] = i;
		tag |= 1 << block;
		block = ((ent->nextblock) & 0xFF) + 1; // ent->nextblock is 0 origin.
	}
	*F_TAG = tag;
	*F_CTRL = 2;

	while(!(*F_CTRL & 1)) { /*nothing*/ }

	// setup internals
	IconIndex = 0;
	CurrentFrame = 0;
	struct schdr *hdr = (struct schdr *)VFLA;
	const u8 *pfvicons = (const u8 *)VFLA
		+ 0x80 // schdr(incl. mcxhdr)
		+ 0x80 * (hdr->numicons & 0x0F) // PS icons
		+ ((hdr->pkmagic & 0x01000000) ? 0x0800 : 0) // snapshot
		+ 8 * 0x10 * ((hdr->numbucommands + 0x0F) / 0x10) // BU funcs
		;
	if(hdr->numexicons == 0) {
		// build ExIconEntry dynamically
		ExIconFromFVIcon.animLength = hdr->numfvicons;
		ExIconFromFVIcon.animDelay = 6; // same as fvicons;
		ExIconFromFVIcon.addr = (u32*)pfvicons;
		// then use it
		ExIconEntries = &ExIconFromFVIcon;
	} else {
		ExIconEntries = (struct exiconentry *)(pfvicons + 0x80 * hdr->numfvicons); // fileviewer icons
	}
	AnimWait = ExIconEntries[0].animDelay;
}

void *scene_app_main(u32 stat, u32 prevstat) {
	PressedSinceSceneEnter &= stat;
	stat &= ~PressedSinceSceneEnter;
	prevstat &= ~PressedSinceSceneEnter;
	u32 press = (stat ^ prevstat) & stat;
	u32 release = (stat ^ prevstat) & prevstat;
	// process key

	if(release & INTMASK_ACTION) {
		runApp(AppIdx2FirstBlock[Selection]);
	} else if(press & INTMASK_LEFT) {
		startBeep();
		if(Selection == 0) {
			return scene_clock_entry;
		}
		Selection--;
		mapApp(Selection);
	} else if(press & INTMASK_RIGHT) {
		startBeep();
		Selection++;
		if(Selection == NumApps) {
			return scene_clock_entry;
		}
		mapApp(Selection);
	} else if(press & INTMASK_UP) {
		struct schdr *hdr = (struct schdr *)VFLA;
		if(hdr->numexicons > 1) {
			startBeep();
			if(IconIndex == 0) {
				IconIndex = hdr->numexicons - 1;
			} else {
				IconIndex--;
			}
			CurrentFrame = 0;
			AnimWait = ExIconEntries[IconIndex].animDelay;
		}
	} else if(press & INTMASK_DOWN) {
		struct schdr *hdr = (struct schdr *)VFLA;
		if(hdr->numexicons > 1) {
			startBeep();
			IconIndex++;
			if(IconIndex == hdr->numexicons) {
				IconIndex = 0;
			}
			CurrentFrame = 0;
			AnimWait = ExIconEntries[IconIndex].animDelay;
		}
	}

	// draw

	for(u32 y = 0; y < 32; y++) {
		OffScreen[y] = ExIconEntries[IconIndex].addr[y];
	}

	// wait or advance frame
	if(1 < AnimWait) {
		AnimWait--;
	} else {
		CurrentFrame++;
		if(ExIconEntries[IconIndex].animLength <= CurrentFrame) {
			CurrentFrame = 0;
		}
		AnimWait = ExIconEntries[IconIndex].animDelay;
	}

	return 0;
}

void *scene_app_entry(u32 stat, u32 prevstat) {
	(void)prevstat;
	if(NumApps == 0) {
		return scene_clock_entry;
	}

	if(stat & INTMASK_LEFT) {
		Selection = NumApps - 1;
	} else {
		Selection = 0;
	}
	if(GuiParamApp) {
		for(u32 i = 0; i < NumApps; i++) {
			if(AppIdx2FirstBlock[i] == GuiParamApp) {
				Selection = i;
				break;
			}
		}
		GuiParamApp = 0;
	}

	mapApp(Selection);

	PressedSinceSceneEnter = stat;

	return scene_app_main;
}
