#include "common_c.h"

extern u32 OffScreen[32];
extern u8 NumUsingBlocks;
extern u8 NumApps;
extern struct alarm_info *PAlarmInfo;
extern s8 GuiParamApp;

static s8 SettingFieldIdx;
static u32 RepeatCount;
static u32 PressCount;
static u32 PressMask;

extern const u32 font4x6[];

u32 calcWeekBCD(const u32 centuryBCD, const u32 year2BCD, const u32 monthBCD, const u32 dayBCD);
void setRTC(const u32 datebcd, const u32 timebcd);
void startBeep();
void clearOffScreen();
void flipScreen();
void runApp(u32 appFirstBlock);
void *scene_app_entry(u32 stat, u32 prevstat);

void drawDateTime(u32 datebcd, u32 timebcd) {
	for(u32 y = 0; y < 5; y++) {
		{
			u32 vl = OffScreen[y];
			vl |= ((font4x6[(datebcd >> 28) & 0x0F] >> (y * 4)) & 0x0F);
			vl |= ((font4x6[(datebcd >> 24) & 0x0F] >> (y * 4)) & 0x0F) << 4;
			vl |= ((font4x6[(datebcd >> 20) & 0x0F] >> (y * 4)) & 0x0F) << 8;
			vl |= ((font4x6[(datebcd >> 16) & 0x0F] >> (y * 4)) & 0x0F) << 12;
			vl |= ((font4x6[(datebcd >> 12) & 0x0F] >> (y * 4)) & 0x0F) << 16;
			vl |= ((font4x6[(datebcd >>  8) & 0x0F] >> (y * 4)) & 0x0F) << 20;
			vl |= ((font4x6[(datebcd >>  4) & 0x0F] >> (y * 4)) & 0x0F) << 24;
			vl |= ((font4x6[(datebcd >>  0) & 0x0F] >> (y * 4)) & 0x0F) << 28;
			OffScreen[y] = vl;
		}

		{
			u32 vl = OffScreen[6+y];
			static const u8 w2ww[8][2] = {{10,21},{22,18},{23,21},{24,11},{23,25},{26,27},{10,28}};
			vl |= ((font4x6[w2ww[(timebcd >> 24) - 1][0]] >> (y * 4)) & 0x0F);
			vl |= ((font4x6[w2ww[(timebcd >> 24) - 1][1]] >> (y * 4)) & 0x0F) << 4;
			vl |= ((font4x6[(timebcd >> 20) & 0x0F] >> (y * 4)) & 0x0F) << 8;
			vl |= ((font4x6[(timebcd >> 16) & 0x0F] >> (y * 4)) & 0x0F) << 12;
			vl |= ((font4x6[(timebcd >> 12) & 0x0F] >> (y * 4)) & 0x0F) << 16;
			vl |= ((font4x6[(timebcd >>  8) & 0x0F] >> (y * 4)) & 0x0F) << 20;
			vl |= ((font4x6[(timebcd >>  4) & 0x0F] >> (y * 4)) & 0x0F) << 24;
			vl |= ((font4x6[(timebcd >>  0) & 0x0F] >> (y * 4)) & 0x0F) << 28;
			OffScreen[6+y] = vl;
		}
	}

	OffScreen[10] |= (*I_STAT & INTMASK_RTC) ? 0x80000000 : 0;
}

void drawAlarm(u32 enabled, u32 hourBCD, u32 minuteBCD) {
	u32 al = enabled ? 34 : 36;
	u32 ar = enabled ? 35 : 37;
	for(u32 y = 0; y < 6/*alarm symbol requires 6 lines*/; y++) {
		u32 vl = OffScreen[26+y];
		vl |= ((font4x6[al] >> (y * 4)) & 0x0F) <<  8;
		vl |= ((font4x6[ar] >> (y * 4)) & 0x0F) << 12;
		vl |= ((font4x6[(hourBCD >> 4) & 0x0F] >> (y * 4)) & 0x0F) << 16;
		vl |= ((font4x6[(hourBCD >> 0) & 0x0F] >> (y * 4)) & 0x0F) << 20;
		vl |= ((font4x6[(minuteBCD >> 4) & 0x0F] >> (y * 4)) & 0x0F) << 24;
		vl |= ((font4x6[(minuteBCD >> 0) & 0x0F] >> (y * 4)) & 0x0F) << 28;
		OffScreen[26+y] = vl;
	}
}

void drawVol(u32 attenate) {
	u32 sl = attenate == 2 ? 30 : 29;
	u32 sr = 31 + (2 - attenate);
	for(u32 y = 0; y < 6/*spk symbol requires 6 lines*/; y++) {
		u32 vl = OffScreen[26+y];
		vl |= ((font4x6[sl] >> (y * 4)) & 0x0F) << 0;
		vl |= ((font4x6[sr] >> (y * 4)) & 0x0F) << 4;
		OffScreen[26+y] = vl;
	}
}

void drawNumUsingBlocks() {
	for(u32 y = 0; y < 5; y++) {
		u32 vl = OffScreen[20+y];
		vl |= ((font4x6[(NumApps >> 4) & 0x0F] >> (y * 4)) & 0x0F) << 0;
		vl |= ((font4x6[(NumApps >> 0) & 0x0F] >> (y * 4)) & 0x0F) << 4;
		vl |= ((font4x6[38] >> (y * 4)) & 0x0F) << 8;
		vl |= ((font4x6[(NumUsingBlocks >> 4) & 0x0F] >> (y * 4)) & 0x0F) << 12;
		vl |= ((font4x6[(NumUsingBlocks >> 0) & 0x0F] >> (y * 4)) & 0x0F) << 16;
		OffScreen[20+y] = vl;
	}
}

void *scene_clock_main(u32 stat, u32 prevstat) {
	static const struct _range settingFieldRange[10] = {{0, 0x99}, {1, 0x12}, {1, 0x31/*dummy*/}, {0, 0x23}, {0, 0x59}, {0, 0x59}, {0, 2}, {0, 1}, {0, 0x23}, {0, 0x59}};
	static u8 settingFields[11]; // c,y2,m,d,h,m,s,attenate,alarm_enable,alarm_h,alarm_m
	static u8 settingWeek; // for cache

	u32 datebcd, timebcd, attenate, alarmena, alarmhbcd, alarmmbcd;

	// initialize process-draw shared vars
	if(SettingFieldIdx == -1) {
		datebcd = sysGetBcdDate();
		timebcd = sysGetBcdTime();
		attenate = PAlarmInfo->attenate;
		alarmena = PAlarmInfo->enabled;
		alarmhbcd = PAlarmInfo->hourBCD;
		alarmmbcd = PAlarmInfo->minuteBCD;
	} else {
		datebcd = (settingFields[0] << 24) | (settingFields[1] << 16) | (settingFields[2] << 8) | (settingFields[3] << 0);
		timebcd = (settingWeek << 24) | (settingFields[4] << 16) | (settingFields[5] << 8) | (settingFields[6] << 0);
		attenate = 2 - settingFields[7];
		alarmena = settingFields[8];
		alarmhbcd = settingFields[9];
		alarmmbcd = settingFields[10];
	}

	// process key

	if(SettingFieldIdx != -1) {
		//u32 changedmask = stat ^ prevmask;
		//u32 pressedmask = stat & changedmask;

		if(SettingFieldIdx == 0) {
			// wait for releasing all buttons, to avoid incidentally modifying.
			if(!(stat & 0x1F)) {
				SettingFieldIdx = 1;
			}
		} else {
			if(stat & 0x1F/*udlr and action*/) {
				RepeatCount++;
			} else {
				RepeatCount= 0;
			}
			if((1 < RepeatCount) && ((RepeatCount< 15) || (((RepeatCount - 15) % 4) != 0))) {
				stat = 0;
			}

			if(stat & INTMASK_UP) {
				const struct _range *range = &settingFieldRange[SettingFieldIdx - 1];
				if(SettingFieldIdx == 1) { // year
					// TODO incr century if 99->00
				} else if(SettingFieldIdx == 3) { // day
					// TODO wrap if month-dependent value
				}
				if(settingFields[SettingFieldIdx] < range->max) {
					settingFields[SettingFieldIdx]++;
					if(9 < (settingFields[SettingFieldIdx] & 0x0F)) {
						settingFields[SettingFieldIdx] += 6;
					}
				} else {
					settingFields[SettingFieldIdx] = range->min; // wrap
				}
				if(SettingFieldIdx <= 3) {
					settingWeek = calcWeekBCD(settingFields[0], settingFields[1], settingFields[2], settingFields[3]);
				}
				if(SettingFieldIdx == 7) {
					PAlarmInfo->attenate = 2 - settingFields[7]; // always apply volume for preview
				}
				startBeep();
			} else if(stat & INTMASK_DOWN) {
				const struct _range *range = &settingFieldRange[SettingFieldIdx - 1];
				if(SettingFieldIdx == 1) { // year
					// TODO decr century if 00->99
				} else if(SettingFieldIdx == 3) { // day
					// TODO wrap to month-dependent value
				}
				if(range->min < settingFields[SettingFieldIdx]) {
					settingFields[SettingFieldIdx]--;
					if(9 < (settingFields[SettingFieldIdx] & 0x0F)) { // wrap
						settingFields[SettingFieldIdx] -= 6;
					}
				} else {
					settingFields[SettingFieldIdx] = range->max; // wrap
				}
				if(SettingFieldIdx <= 3) {
					settingWeek = calcWeekBCD(settingFields[0], settingFields[1], settingFields[2], settingFields[3]);
				}
				if(SettingFieldIdx == 7) {
					PAlarmInfo->attenate = 2 - settingFields[7]; // always apply volume for preview
				}
				startBeep();
			} else if(stat & INTMASK_LEFT) {
				if(1 < SettingFieldIdx) {
					SettingFieldIdx--;
				} else {
					SettingFieldIdx = 10;
				}
				startBeep();
			} else if(stat & INTMASK_RIGHT) {
				if(SettingFieldIdx < 10) {
					SettingFieldIdx++;
				} else {
					SettingFieldIdx = 1;
				}
				startBeep();
			} else if(stat & INTMASK_ACTION) {
				SettingFieldIdx = -1;
				setRTC(datebcd, timebcd);
				PAlarmInfo->attenate  = attenate;
				PAlarmInfo->enabled   = alarmena;
				PAlarmInfo->hourBCD   = alarmhbcd;
				PAlarmInfo->minuteBCD = alarmmbcd;
				if(GuiParamApp) {
					runApp(GuiParamApp);
				} else {
					startBeep();
				}
			}
		}
	} else {
		u32 press = (stat ^ prevstat) & stat;
		if(stat & INTMASK_UP) {
			PressMask = INTMASK_UP;
			PressCount++;
			if(PressCount == 30) {
				PressMask = 0;
				PressCount = 0;
				settingFields[0] = (datebcd >> 24) & 0xFF;
				settingFields[1] = (datebcd >> 16) & 0xFF;
				settingFields[2] = (datebcd >>  8) & 0xFF;
				settingFields[3] = (datebcd >>  0) & 0xFF;
				settingFields[4] = (timebcd >> 16) & 0xFF;
				settingFields[5] = (timebcd >>  8) & 0xFF;
				settingFields[6] = (timebcd >>  0) & 0xFF;
				settingFields[7] = 2 - PAlarmInfo->attenate;
				settingFields[8] = PAlarmInfo->enabled;
				settingFields[9] = PAlarmInfo->hourBCD;
				settingFields[10] = PAlarmInfo->minuteBCD;
				SettingFieldIdx = 0;
				settingWeek = calcWeekBCD(settingFields[0], settingFields[1], settingFields[2], settingFields[3]);
				startBeep();
			}
		} else if(stat & INTMASK_DOWN) {
			PressMask = INTMASK_DOWN;
			PressCount++;
		} else if((press & INTMASK_LEFT) || (press & INTMASK_RIGHT)) {
			startBeep();
			return scene_app_entry;
		} else {
			PressMask = 0;
			PressCount = 0;
		}
	}

	// draw

	if(!(PressMask & INTMASK_UP) || PressCount % 8 < 4) {
		drawDateTime(datebcd, timebcd);
		drawVol(attenate);
		drawAlarm(alarmena, alarmhbcd, alarmmbcd);
	}
	if(!(PressMask & INTMASK_DOWN) || PressCount % 8 < 4) {
		drawNumUsingBlocks();
	}

	if(0 < SettingFieldIdx) {
		struct _area {
			struct _range x, y;
		};
		static const struct _area areas[10] = {
			{{4*0, 4*4}, { 0,  6}},
			{{4*4, 4*6}, { 0,  6}},
			{{4*6, 4*8}, { 0,  6}},
			{{4*2, 4*4}, { 6, 12}},
			{{4*4, 4*6}, { 6, 12}},
			{{4*6, 4*8}, { 6, 12}},
			{{4*0, 4*2}, {26, 32}},
			{{  9, 4*4}, {26, 32}},
			{{4*4, 4*6}, {26, 32}},
			{{4*6, 4*8}, {26, 32}},
		};
		const struct _area *area = &areas[SettingFieldIdx - 1];
		const u32 mask = ((1 << (area->x.max - area->x.min)) - 1) << area->x.min;
		for(u32 y = area->y.min; y < area->y.max; y++) {
			OffScreen[y] ^= mask;
		}
	}

	return 0;
}

void *scene_clock_entry(u32 stat, u32 prevstat) {
	(void)stat, (void)prevstat;
	if(GuiParamApp) {
		SettingFieldIdx = 0;
		if(GuiParamApp < 0) {
			GuiParamApp = 0;
		}
	} else {
		SettingFieldIdx = -1;
	}
	RepeatCount = 0;
	PressCount = 0;
	PressMask = 0;
	return scene_clock_main;
}
