#include "common_c.h"

extern u32 svcFlashWritePhysical(u32 sector, void *ptr);
extern u32 svcFlashReadSerial();
extern u32 svcGetBcdDate();
extern u32 svcGetBcdTime();
extern void svcSetBcdDateTime(u32 bcd_date, u32 bcd_weektime);
extern s32 getMCXType(u32 diridx);

struct bufunc1ret *sysbufunc0(u32 flags, void *buf);
struct bufunc1ret *sysbufunc1(u32 flags, void *buf);
struct bufunc1ret *sysbufunc2(u32 flags, void *buf);
struct bufunc1ret *sysbufunc3(u32 flags, void *buf);

static struct bufuncentry sysBUFuncs[] = {
	{0, sysbufunc0},
	{5, sysbufunc1},
	{0, sysbufunc2},
	{4, sysbufunc3},
};

static int isThisSectorCurrentAppDirEntry(u32 sector) {
	if(CurrentExecAppIndex == 0) {
		// GUI is special
		return 0;
	}
	if(sector <= 0 || 15 < sector) {
		// dirindex is out of range
		return 0;
	}
	u32 block = CurrentExecAppIndex;
	for(;;) {
		if(block == sector) {
			byteF0 = 8;
			byteF1 = 1;
			return 1;
		}
		struct mcdir *pent = (struct mcdir *)((u8*)PFLA + block * 0x80);
		if(pent->nextblock == 0xFFFF) {
			return 0;
		}
		block = pent->nextblock;
	}
	return 0;
}

static struct bufuncentry *getPtrToBUFuncTab() {
	struct schdr *schdr = (struct schdr *)0x02000000; // VFLA
	return (struct bufuncentry *)((u8*)schdr + ((schdr->numicons & 0x0F) + 1) * 0x80 + ((getMCXType(CurrentExecAppIndex) == 1) ? 0x800 : 0));
}

int waitComEnd() {
	for(;;) {
		if(*COM_STAT1 & 0x00000002) {
			return 0;
		}
		if(!(*I_STAT & INTMASK_IOP)) {
			return -1;
		}
	}
}

int waitComTxComplete() {
	for(;;) {
		if(!(*I_STAT & INTMASK_IOP)) {
			return -1;
		}
		if(*COM_STAT1 & 0x00000002) {
			return -1;
		}
		if(*COM_STAT2 & 0x00000001) {
			return 0;
		}
	}
}

int ctrlAndWaitComTxComplete() {
	*COM_CTRL2 = 1;

	int r = waitComTxComplete();

	if(!r) {
		*COM_CTRL2 = 1;
	}

	return r;
}

int comTxByte(u32 byte) {
	*COM_CTRL2 = 1;
	*COM_DATA = byte;
	*COM_MODE = 1;
	*COM_MODE = 3;

	for(;;) {
		if(!(*I_STAT & INTMASK_IOP)) {
			break;
		}
		if(*COM_STAT1 & 0x00000002) {
			break;
		}
		if(*COM_STAT2 & 0x00000001) {
			*COM_CTRL2 = 1;
			*COM_STAT1; // empty read
			return 0;
		}
	}

	*COM_CTRL2 = 1;
	*COM_STAT1; // empty read
	return 1;
}

int comTxByteAndWaitEnd(u32 byte) {
	*COM_DATA = byte;
	*COM_MODE = 1;
	*COM_MODE = 3;
	return waitComEnd();
}

// get/set RTC
struct bufunc1ret *sysbufunc0(u32 flags, void *buf) {
	(void)buf; // unused

	if(flags & BUFUNC_ERROR) {
		return 0;
	}

	BUBuf.bufSize = 8;
	BUBuf.pBuf = SectBuf;

	if((flags & BUFUNC_TOPS)) {
		if(flags & BUFUNC_STAGE2) {
			// PK2PS, Stage2
			return 0;
		}
		// PK2PS, Stage1
		u32 *p = (u32*)BUBuf.pBuf;
		p[0] = svcGetBcdDate();
		p[1] = svcGetBcdTime();
		return &BUBuf;
	} else {
		// FromPS
		//if((flags & BUFUNC_FROMPS) && !(flags & BUFUNC_STAGE1)/*flags & BUFUNC_STAGE2?*/) {
		if(flags & BUFUNC_STAGE2) {
			// FromPS, Stage2
			KernFlags |= KERNFLAGS_SETRTC;
		}
		// FromPS, Stage1/2
		return &BUBuf;
	}
}

// bulk read/write pk memory
struct bufunc1ret *sysbufunc1(u32 flags, void *buf) {
	if(flags & BUFUNC_ERROR) {
		return 0;
	}

	if(flags & BUFUNC_STAGE1) {
		BUBuf.bufSize = ((u8*)buf)[4];
		if(flags & BUFUNC_TOPS) {
			// tops, stage1
			BUBuf.pBuf = ((void**)buf)[0]; // directly set source as buffer.
		} else if(flags & BUFUNC_FROMPS) {
			// fromps, stage1
			BUBuf.pBuf = &SectBuf;
			Sysbufunc1DstPtr = ((void**)buf)[0]; // save dstaddr
		}
		return &BUBuf;
	}

	if(flags & BUFUNC_TOPS) {
		// tops, stage2?
		return 0;
	}
	if(!(flags & BUFUNC_STAGE2)) {
		// fromps?, stage1? never reach here?
		return 0;
	}
	// fromps? stage2.
	void *psrc = BUBuf.pBuf;
	void *pdst = Sysbufunc1DstPtr; // may or may not be aligned
#if 0
	// original: pdst align is ok but psrc align will be broken!!
	while(BUBuf.bufSize) {
		if(((u32)psrc & 1) || (BUBuf.bufSize == 1) || (BUBuf.bufSize == 3)) {
			*(u8*)pdst++ = *(u8*)psrc++;
			BUBuf.bufSize -= 1;
		} else if(((u32)psrc & 2) || (BUBuf.bufSize == 2)) {
			*(u16*)pdst++ = *(u16*)psrc++;
			BUBuf.bufSize -= 2;
		} else {
			*(u32*)pdst++ = *(u32*)psrc++;
			BUBuf.bufSize -= 4;
		}
	}
#else
	// slow but reliable
	for(u32 i = 0; i < BUBuf.bufSize; i++) {
		*(u8*)pdst++ = *(u8*)psrc++;
	}
#endif
	return 0;
}

// get/set alarm setting
struct bufunc1ret *sysbufunc2(u32 flags, void *buf) {
	(void)buf; // unused

	if(flags & BUFUNC_ERROR) {
		return 0;
	}

	if(!(flags & BUFUNC_STAGE1)) {
		return 0;
	}

	// stage1
	BUBuf.pBuf = AlarmSettings;
	BUBuf.bufSize = sizeof(AlarmSettings); // 8
	return &BUBuf;
}

// call some function... but I can't get its purpose... it seems almost broken...!!
struct bufunc1ret *sysbufunc3(u32 flags, void *buf) {
	if(flags & BUFUNC_ERROR) {
		return 0;
	}

	if(BUFunc3offset == 0) {
		*(u32*)4 = 0; // TODO WTF???
		return 0;
	}

	// TODO argcount is unknown. seems 1(only touched r0) or 2(r1 is passthrough?) or 3(use r3 as funcptr=r0..r2 is arg?).
	return ((struct bufunc1ret *(*)(u32, void*, u32))(0x02000000/*VFLA*/ + BUFunc3offset))(flags, buf, flags);
}

// TODO should psrcblk be "volatile u16*" to force access by halfword?
int sendFlashBlock(const u16 *psrcblk, u32 *pcksum) {
	u16 hwbuf = hwbuf; // FIXME self-initializing to disable -fmaybe-uninitialized for this

	for(u32 i = 0; i < 0x80; i++) {
		u8 b;
		if(i % 2 == 0) {
			hwbuf = *psrcblk++;
			b = hwbuf & 0xFF;
		} else {
			b = (hwbuf >> 8) & 0xFF;
		}

		*COM_CTRL2 = 1;
		*pcksum ^= b;
		*COM_DATA = b;
		*COM_MODE = 1;
		*COM_MODE = 3;

		if(waitComTxComplete()) { return 1; }
	}

	*COM_CTRL2 = 1;

	return 0;
}

void commReadBottomHalf() {
	if(comTxByte(0x00)) { return; }
	u8 r1 = *COM_DATA;
	if(comTxByte(0x00)) { return; }
	u8 r2 = *COM_DATA;
	if(comTxByte(0x5C)) { return; } // backslash
	// memo: no$ says this wait is compatibility with original Sony MC
	for(u32 i = 251; 0 < i; i--) { /*nothing*/ } // original: i=250;while(i--!=0)/*nothing*/;
	if(comTxByte(0x5D)) { return; } // ]

	u16 sector = (r1 << 8) | r2;
	if(0x400 <= sector) { // out of range
		if(comTxByte(0xFF)) { return; }
		if(comTxByte(0xFF)) { return; } // original: no error check (optimized?)
		return;
	}

	if(comTxByte(r1)) { return; }
	if(comTxByte(r2)) { return; }

	const u16 *psrc = (const u16*)((const u8*)PFLA + (sector * 0x80));
	u32 cksum = r1 ^ r2; // checksum including first 2 response bytes
	if(sendFlashBlock(psrc, &cksum)) { return; }
	if(comTxByte(cksum)) { return; }
	comTxByteAndWaitEnd(0x47); // G
}

int recvBlock(u8 *pdst, u32 *pcksum) {
	for(u32 i = 0; i < 0x80; i++) {
		u8 data = *COM_DATA;
		*pdst++ = data;
		*pcksum ^= data;
		*COM_DATA = 0;
		*COM_MODE = 1;
		*COM_MODE = 3;

		u32 comstat1;
		for(;;) {
			if(!(*I_STAT & INTMASK_IOP)) {
				return 1;
			}
			comstat1 = *COM_STAT1;
			if(comstat1 & 0x00000003) {
				break;
			}
		}
		if(comstat1 & 0x00000002) {
			return 1;
		}
	}
	*COM_CTRL2 = 1;
	return 0;
}

void commWriteBottomHalf() {
	if(comTxByte(0x00)) { return; }
	u8 r1 = *COM_DATA;
	if(comTxByte(0x00)) { return; }
	u8 r2 = *COM_DATA;
	if(comTxByte(0x00)) { return; }

	u32 cksum = r1 ^ r2;
	if(recvBlock(SectBuf, &cksum)) { return; }

	u8 r3 = *COM_DATA;
	if(comTxByte(0x5C)) { return; } // backslash
	if(comTxByte(0x5D)) { return; } // ]

	u8 ret = 0x47; // G: OK
	if(r3 != cksum) {
		ret = 0x4E; // N: NG
	}
	u16 sector = (r1 << 8) | r2;
	if(0x400 <= sector) { // out of range
		ret = 0xFF;
	}
	if(isThisSectorCurrentAppDirEntry(sector)) {
		ret = 0xFE;
	}
	if((KernFlags & KERNFLAGS_WPALT) && (between(16, sector, 35)/*alt sector list*/ || between(36, sector, 55)/*alt sector data*/)) {
		ret = 0xFD;
	}

	comTxByteAndWaitEnd(ret);

	if(ret == 0x47) {
		// if good, complete COM, then write it to Flash.
		*COM_DATA = 0xFF;
		*COM_CTRL1 = 0;
		*COM_MODE = 2;
		if(!svcFlashWritePhysical(sector, SectBuf)) {
			// write succeed
			byteF0 = 0;
		} else {
			// write failed?
			byteF0 = 4;
		}
	}
}

void commSBottomHalf() {
	if(comTxByte(0x5C)) { return; } // backslash
	if(comTxByte(0x5D)) { return; } // ]

	// memo: no$ says it might be 0400,0080 which is total-sectors and sector-size?
	if(comTxByte(0x04)) { return; }
	if(comTxByte(0x00)) { return; }

	if(comTxByte(0x00)) { return; }
	comTxByteAndWaitEnd(0x80);
}

void comm50() {
	if(!comTxByte(0x00)) { return; }
	//byteCA = *COM_DATA; // byteCA is never read; skipping write.
	*COM_DATA; // dummy read
}

void comm58() {
	if(comTxByte(0x02)) { return; }
	if(comTxByte(0x01)) { return; }
	if(comTxByte(0x01)) { return; }
}

void commPrepareExecute() {
	if(comTxByte(0x06)) { return; }
	if(comTxByte(CurrentExecAppIndex >> 8)) { return; }
	u8 r1 = *COM_DATA;
	if(comTxByte(CurrentExecAppIndex & 0xFF)) { return; }
	u8 r2 = *COM_DATA;
	u16 r12 = (r1 << 8) | r2;

	if(comTxByte(0x00)) { return; }
	u8 r3 = *COM_DATA;
	if(comTxByte(0x00)) { return; }
	u8 r4 = *COM_DATA;
	if(comTxByte(0x00)) { return; }
	u8 r5 = *COM_DATA;
	if(comTxByte(0x00)) { return; }
	u8 r6 = *COM_DATA;
	u16 r6543 = (r6 << 24) | (r5 << 16) | (r4 << 8) | r3; // yes, reverse order!

	if(r12 == 0xFFFF) {
		return;
	}
	if(r12 == 0xFFFE) {
		KernFlags |= KERNFLAGS_REBOOT;
	} else if(r12 <= 15) {
		PreparedAppIndex = r12;
		PreparedAppParam = r6543;
		KernFlags |= KERNFLAGS_PREPARED; // Kernel only sets this flag, no handling. App should do that.
	}
}

void commSendPkStatus() {
	if(comTxByte(0x12)) { return; }

	if(comTxByte(CurrentExecAppIndex >> 8)) { return; }
	if(comTxByte(CurrentExecAppIndex & 0xFF)) { return; }

	if(comTxByte((KernFlags & KERNFLAGS_0001) ? 1 : 0)) { return; }
	if(comTxByte((KernFlags & KERNFLAGS_0002) ? 1 : 0)) { return; }
	if(comTxByte((KernFlags & KERNFLAGS_0008) ? 1 : 0)) { return; } // yes, 8.
	if(comTxByte((KernFlags & KERNFLAGS_0004) ? 1 : 0)) { return; } // yes, 4.

	u32 sn = svcFlashReadSerial();
	if(comTxByte((sn >>  0) & 0xFF)) { return; }
	if(comTxByte((sn >>  8) & 0xFF)) { return; }
	if(comTxByte((sn >> 16) & 0xFF)) { return; }
	if(comTxByte((sn >> 24) & 0xFF)) { return; }

	u32 date = svcGetBcdDate();
	if(comTxByte((date >>  0) & 0xFF)) { return; }
	if(comTxByte((date >>  8) & 0xFF)) { return; }
	if(comTxByte((date >> 16) & 0xFF)) { return; }
	if(comTxByte((date >> 24) & 0xFF)) { return; }

	u32 time = svcGetBcdTime();
	if(comTxByte((time >>  0) & 0xFF)) { return; }
	if(comTxByte((time >>  8) & 0xFF)) { return; }
	if(comTxByte((time >> 16) & 0xFF)) { return; }
	if(comTxByte((time >> 24) & 0xFF)) { return; } // original: no error check; optimized?
}

static u32 getBUFuncParamLen(u32 funcNo) {
	if(funcNo < 0x80) {
		return sysBUFuncs[funcNo].paramLen;
	} else {
		// 0x80 <= funcNo
		return getPtrToBUFuncTab()[funcNo - 0x80].paramLen;
	}
}

static BUFunc *getBUFuncPtr(u32 funcNo) {
	if(funcNo < 0x80) {
		return sysBUFuncs[funcNo].pFunc;
	} else {
		// 0x80 <= funcNo
		return getPtrToBUFuncTab()[funcNo - 0x80].pFunc;
	}
}

u32 recvBUFuncNoAndParam() {
	if(comTxByte(0xFF)) { return -1; }
	u8 r1 = *COM_DATA;
	if(CurrentExecAppIndex == 0 && 4 <= r1) { return -1; }

	u32 paramlen = getBUFuncParamLen(r1);
	if(comTxByte(paramlen)) { return -1; }

	u8 *p = SectBuf;
	for(u32 i = 0; i < paramlen; i++) {
		if(comTxByte(0x00)) { return -1; }
		*p = *COM_DATA;
	}

	return r1;
}

int BUFuncPrelude(BUFunc **ppfunc, u32 f1flag, struct bufunc1ret **pfret, u32 *plen2) {
	u32 funcNo = recvBUFuncNoAndParam();
	if(0x100u <= funcNo) { return 1; }
	*ppfunc = getBUFuncPtr(funcNo);
	if(!*ppfunc) { return 1; }
	*pfret = (*ppfunc)(f1flag, SectBuf);
	if(*pfret == 0) { return 1; }
	*plen2 = ((*pfret)->bufSize < 0x80) ? (*pfret)->bufSize : 0x80;
	if(!comTxByte(*plen2)) { return 1; }
	return 0;
}

void commBUFuncSendToPS() {
	BUFunc *pfunc;
	struct bufunc1ret *fret;
	u32 len2;
	if(BUFuncPrelude(&pfunc, BUFUNC_STAGE1 | BUFUNC_TOPS, &fret, &len2)) { return; }

	u32 f2flag = BUFUNC_STAGE2 | BUFUNC_TOPS;
	for(u32 i = 0; i < len2; i++) {
		// reading as halfword; considering read from Flash? (but is it really required?)
		u32 addr = (u32)fret->pBuf + i;
		const u16 *psrc = (const u16 *)(addr & ~1); // halfword align
		u8 b = *psrc >> ((addr & 1) * 8);
		if(!comTxByte(b)) {
			f2flag |= BUFUNC_ERROR;
			break;
		}
	}
	*COM_DATA = 0xFF;
	*COM_MODE = 1;
	*COM_MODE = 3;
	pfunc(f2flag, fret->pBuf);
}

void commBUFuncRecvFromPS() {
	BUFunc *pfunc;
	struct bufunc1ret *fret;
	u32 len2;
	if(BUFuncPrelude(&pfunc, BUFUNC_STAGE1 | BUFUNC_FROMPS, &fret, &len2)) { return; }

	u32 f2flag = BUFUNC_STAGE2 | BUFUNC_FROMPS;
	for(u32 i = 0; i < len2; i++) {
		if(!comTxByte(0x00)) {
			f2flag |= BUFUNC_ERROR;
			break;
		}
		((u8*)fret->pBuf)[i] = *COM_DATA;
	}
	*COM_DATA = 0xFF;
	*COM_MODE = 1;
	*COM_MODE = 3;
	pfunc(f2flag, fret->pBuf);
}

// no$ says this is "Execute Custom Download Notification Function"
void commCallCustom3() {
	if(!comTxByte(0x03)) { return; }
	if(!comTxByte(0x00)) { return; }
	u8 r1 = *COM_DATA;
	if(!comTxByte(0x00)) { return; }
	u8 r2 = *COM_DATA;
	if(!comTxByte(0x00)) { return; }
	u8 r3 = *COM_DATA;

	// wtf...
	if(r2 == 0) {
		KernFlags |= KERNFLAGS_WPALT;
	} else {
		KernFlags &= ~KERNFLAGS_WPALT;
	}

	if(CustomHandlers[3] == 0) { return; }

	((void(*)(u32))CustomHandlers[3])((r1 << 16) | (r2 << 8) | r3);
}

void comm5E() {
	if(!comTxByte(0x03)) { return; }

	if(!comTxByte((KernFlags & KERNFLAGS_0002) ? 1 : 0)) { return; }
	{
		u8 r1 = *COM_DATA;
		if(r1 == 1) {
			KernFlags |= KERNFLAGS_0002;
		} else if(r1 == 0) {
			KernFlags &= ~KERNFLAGS_0002;
		}
	}

	if(!comTxByte((KernFlags & KERNFLAGS_0008) ? 1 : 0)) { return; }
	{
		u8 r2 = *COM_DATA;
		if(r2 == 1) {
			KernFlags |= KERNFLAGS_0008;
		} else if(r2 == 0) {
			KernFlags &= ~KERNFLAGS_0008;
		}
	}

	if(!comTxByte((KernFlags & KERNFLAGS_0004) ? 1 : 0)) { return; }
	{
		u8 r3 = *COM_DATA;
		if(r3 == 1) {
			KernFlags |= KERNFLAGS_0004;
		} else if(r3 == 0) {
			KernFlags &= ~KERNFLAGS_0004;
		}
	}
}

void comm5F() {
	if(!comTxByte(0x01)) { return; }

	if(!comTxByte((KernFlags & KERNFLAGS_0001) ? 1 : 0)) { return; }
	{
		u8 r1 = *COM_DATA;
		if(r1 == 1) {
			KernFlags |= KERNFLAGS_0001;
		} else if(r1 == 0) {
			KernFlags &= ~KERNFLAGS_0001;
		}
	}
}

void processComm() {
	*COM_CTRL2 = 1;
	*COM_CTRL1 = 3;

	if(!(KernFlags & KERNFLAGS_COMEN)) {
		return;
	}

	u32 i;

	for(i = 30; 0 < i; i--) {
		if(!(*I_STAT & INTMASK_IOP)) {
			break;
		}
		if(*COM_STAT1 & 0x00000002) {
			continue;
		}
		if(!(*COM_STAT2 & 0x00000001)) {
			continue;
		}

		*COM_CTRL2 = 1;
		break;
	}

	if(0 < i) {
		if(*COM_DATA == 0x81) {
			if(byteF1 != 0) {
				byteF1--;
				*COM_DATA = 0xFF;
				*COM_CTRL1 = 0;
				*COM_MODE = 2;
				waitComEnd();
			} else {
				*COM_DATA = byteF0;
				*COM_MODE = 1;
				*COM_MODE = 3;
				if(byteF0 == 4) {
					byteF0 = 0;
				}
				if(!ctrlAndWaitComTxComplete()) {
					u32 data = *COM_DATA;
#if 0
					switch(data) {
						case 0x50: comm50(); break; // P
						case 0x52: // R
						case 0x57: // W
						case 0x53: // S
							if(!comTxByte(0x5A)) { // Z
								if(!comTxByte(0x5D)) { // ]
									KernFlags |= KERNFLAGS_MCRW;
									switch(data) {
										case 0x52: commReadBottomHalf(); break; // R
										case 0x57: commWriteBottomHalf(); break; // W
										case 0x53: commSBottomHalf(); break; // S
									}
								}
							}
							break;
						case 0x58: comm58(); break; // X
						case 0x59: commPrepareExecute(); break; // Y
						case 0x5A: commSendPkStatus(); break; // Z
						case 0x5B: commBUFuncSendToPS(); break; // [
						case 0x5C: commBUFuncRecvFromPS(); break; // backslash
						case 0x5D: commCallCustom3(); break; // ]
						case 0x5E: comm5E(); break; // ^
						case 0x5F: comm5F(); break; // _
						// no default:
					}
#else
					if(data == 0x50) { comm50(); } // P
					else if(data == 0x52 || data == 0x57 || data == 0x53) { // R,W,S
						if(!comTxByte(0x5A)) { // Z
							if(!comTxByte(0x5D)) { // ]
								KernFlags |= KERNFLAGS_MCRW;
								switch(data) {
									case 0x52: commReadBottomHalf(); break; // R
									case 0x57: commWriteBottomHalf(); break; // W
									case 0x53: commSBottomHalf(); break; // S
								}
							}
						}
					} else if(data == 0x58) { comm58(); } // X
					else if(data == 0x59) { commPrepareExecute(); } // Y
					else if(data == 0x5A) { commSendPkStatus(); } // Z
					else if(data == 0x5B) { commBUFuncSendToPS(); } // [
					else if(data == 0x5C) { commBUFuncRecvFromPS(); } // backslash
					else if(data == 0x5D) { commCallCustom3(); } // ]
					else if(data == 0x5E) { comm5E(); } // ^
					else if(data == 0x5F) { comm5F(); } // _
#endif
					// tail; XXX why do it here? not in sysbufunc0??
					if(KernFlags & KERNFLAGS_SETRTC) {
						KernFlags &= ~KERNFLAGS_SETRTC;
						//svcSetBcdDateTime(*((u32*)SectBuf), *((u32*)SectBuf + 1));
						//union { u8 u8[0x80]; u32 u32[0x20]; } *sectBuf = (void*)&SectBuf;
						//svcSetBcdDateTime(sectBuf->u32[0], sectBuf->u32[1]);
						u32 *sectBuf = (void*)&SectBuf[0];
						svcSetBcdDateTime(sectBuf[0], sectBuf[1]);
					}
				}
			}
		}
		// commonexit
		*COM_MODE = 3;
		*COM_STAT1; // dummy read
		*COM_CTRL1 = 0;
		*COM_CTRL2 = 3;
		*COM_MODE = 6;
		*COM_DATA = 0xFF;
		*I_HCLR = INTMASK_COM;
		*COM_CTRL1 = 2;
	}
}
