#include "common_S.h"

public_func panic
	.arm
	b .

private_func excSvc
	.arm
	// All svc handler take at most 3 args(r0-r2) except SVC 02 (user-defined).
	// In original kernel, r7,r11,r12 are used internally.
	// This code uses r12 and lr for working reg.
	// In APCS compliant code, r4-r11 is saved.
	// r0 may be changed for returning value, thus no need to save.
	// so only r1-r3, r12 and lr must be saved.
	// note: if you modify this reg-list, modify svcDoExecute too!!
	stmfd sp!, {r1-r3, r12, lr}
	// get previous insn (that must be SVC)
	mrs r12, SPSR
	tst r12, #PSR_THUMB
	subeq lr, #2 // -2 more if ARM mode (T-bit is 0)
	sub   lr, #2 // -2
	ldrh r12, [lr]
	and r12, #0xFF // filter to get SVC number

	.reloc ., R_ARM_ABS8, PtrToSvcVector
	mov lr, #0
	ldr lr, [lr]

	// We can't assume max SVC index is 0x18; user can override whole table.
	// We can compare *PtrToSvcVector == defaultSvcVector, but it costs.
	//cmp r12, #0x18
	//bhi panic

	ldr r12, [lr, r12, lsl #2]

	mov lr, pc
	bx r12

	ldmfd sp!, {r1-r3, r12, pc}^

private_func excIrq
	.arm
	sub lr, #4 // ARM/thumb common.
	// preserve APCS working regs and LR. use r4 for work-var
	// note: original only preserved r0,r1,r12 and lr. do not assume r2-r4 are saved, on user IRQ handler!
	stmfd sp!, {r0-r3, r4, r12, lr}

	// call user handler first
	.reloc ., R_ARM_ABS8, CustomHandlers // +4
	mov r0, #4
	ldr r0, [r0]
	cmp r0, #0
	movne lr, pc
	bxne r0

	// kernel's own handling follows.

	// call initDockRelatedDevices when docked (INTMASK_IOP has held and mask for that is enabled).
	mov r0, #0x0A000000
	ldr r1, [r0] // I_HOLD
	ldr r4, [r0, #0x08] // I_MASK, use r4(one of save-regs) to avoid destroying by initDockRelatedDevices.
	and r4, r1 // I_HOLD & I_MASK
	tst r4, #INTMASK_IOP
	ldrne r0, =initDockRelatedDevices
	movne lr, pc
	bxne r0 // I_HCLR=INTMASK_IOP in initDockRelatedDevices

	// handle RTC (for century handling)
	tst r4, #INTMASK_RTC
	ldrne r0, =handleRTCIRQ
	movne lr, pc
	bxne r0 // I_HCLR=INTMASK_RTC in handleRTCIRQ

	ldmfd sp!, {r0-r3, r4, r12, pc}^

private_func excFiq
	.arm
	// FIQ: r8-r12, sp, lr are banked(saved).
	// in APCS r0-r3 and r12(IP) are not preserved.
	// FIQ preserves r12_usr, so only r0-r3 must be preserved (ex. by pushing before calling C routine)
	// we use r8, r10 and r11 for work (they are preserved by FIQ, so we can use without pushing.)
	// r9 is... unused. just because.
	sub r8, lr, #4 // ARM/thumb common.

	// kernel COM handling
	mov r11, #0x0A000000
	ldr r10, [r11, #0] // I_HOLD
	ldr r11, [r11, #8] // I_MASK
	and r10, r11
	tst r10, #INTMASK_COM
	bne handleFiqCom // may or may not return

1:
	.reloc ., R_ARM_ABS8, CustomHandlers // +8
	mov r10, #8
	ldr r10, [r10]
	cmp r10, #0
	moveqs pc, r8 // note: return here if r10 is 0. not jump to r10.

	// preserve regs as original code does.
	// note: original preserves r0,r1,r12 and lr. it does not preserve r2-r3.
	//       because r12 is banked by FIQ, storing it is not required.
	//       lr is saved to r8, and r8 is preserved in APCS, it seems storing it is not required...
	//       but... user can use r8 without preserving because original kernel does not use r8 and FIQ saves it?
	//       ok, preserve r8 here.
	//       do not assume that APCS working regs(r0-r3,r12,lr; esp. r2-r3) are preserved by kernel, when programming gui/app!
	stmfd sp!, {r0-r1, r8}

	mov lr, pc
	bx r10

	ldmfd sp!, {r0-r1, r8}
	movs pc, r8

handleFiqCom:
	// preserve APCS work regs in r0-r7 (=r0-r3) that is not banked by FIQ.
	// r8(=lr_fiq) is guaranteed to preserved by callee (APCS function in kernel, that is controlled by me).
	stmfd sp!, {r0-r3}
	ldr r0, =processComm
	mov lr, pc
	bx r0

	.reloc ., R_ARM_ABS8, KernFlags
	mov r10, #0
	ldr r11, [r10]
	mov r12, #KERNFLAGS_REBOOT
	tst r11, r12
	// if not set reboot flag, continue fiq.
	ldmeqfd sp!, {r0-r3}
	beq 1b

	// restart flag is set; reboot to GUI. noreturn.

	// clear flag (reusing r10-r12)
	bic r11, r12
	str r11, [r10]

	// reusing register is over. and no return here. so breaking all registers is ok.

	// run gui
	mov r0, #0
	.reloc ., R_ARM_ABS8, CurrentExecAppIndex
	mov r1, #0
	strh r0, [r1]

	mov r1, #0x0A000000
	mov r0, #INTMASK_COM | INTMASK_RTC | INTMASK_IOP
	str r0, [r1, #8] // I_MASK
	mvn r0, r0
	str r1, [r1, #0x0C] // I_MCLR
	str r1, [r1, #0x10] // I_HCLR

	// breaking lr is ok
	bl clearCustomHandlers

	// reset SV sp
	mov sp, #200

	// restart GUI by jumping bottom of initialize.
	// reboot by COM also executing GUI with param=0x00.
	b restartGui

	.pool

public_rodata armVectorSrc
	.arm
	// note: some production app touches this table (at 0x000) directly... do not modify layout!
	ldr pc, [pc, #0x18] // reset
	ldr pc, [pc, #0x18] // undefined
	ldr pc, [pc, #0x18] // svc
	ldr pc, [pc, #0x18] // prefetch
	ldr pc, [pc, #0x18] // abort
	ldr pc, [pc, #0x18] // reserved; should be NOP?
	ldr pc, [pc, #0x18] // irq
	ldr pc, [pc, #0x18] // fiq
	.word start    // reset
	.word panic    // undefined
	.word excSvc   // svc
	.word panic    // prefetch
	.word panic    // abort
	.word panic    // reserved
	.word excIrq   // irq
	.word excFiq   // fiq
