#include "common_c.h"

void initDockRelatedDevices() {
	if(KernFlags & KERNFLAGS_AUTODAC) {
		*DAC_CTRL &= ~DACCTRL_ENABLE;
		*IOP_STOP = IOPPIN_SOUND;
	}

	if(KernFlags & KERNFLAGS_AUTOIR) {
		*IRDA_MODE |= IRDAMODE_DISABLE;
	}

	*IOP_STOP = IOPPIN_REDLED;

	u32 wait = *CLK_MODE & 0x0F;
	if(wait & 0x08) { wait = 8; }
	wait = 0x20 << wait;
	if(*F_WAIT1 & 0x10) { wait /= 2; }
	while(0 < --wait) { /*nothing*/ }

	// clear hold INTMASK_IOP to avoid re-calling initDockRelatedDevices.
	*I_HCLR = INTMASK_IOP;

	KernFlags |= KERNFLAGS_SENSERQ;
	KernFlags &= ~KERNFLAGS_WPALT;

	u32 docked = *I_STAT & INTMASK_IOP;
	if(docked) {
		KernFlags |= KERNFLAGS_0001 | KERNFLAGS_0002 | KERNFLAGS_0004 | KERNFLAGS_0008;
	} else {
		KernFlags &= ~(KERNFLAGS_0001 | KERNFLAGS_0002 | KERNFLAGS_0004 | KERNFLAGS_0008);
	}

	if(KernFlags & KERNFLAGS_AUTOLCD) {
		if(docked) {
			*LCD_CTRL |= LCDCTRL_ROTATE;
		} else {
			*LCD_CTRL &= ~LCDCTRL_ROTATE;
		}
	}

	byteF0 = 0x08;
}

void initDeviceRegisters() {
	*I_MCLR = 0x3FFF; // all INTMASK, with unknown 0x0020

	*RTC_MODE = RTCMODE_RUN | RTCMODE_UNSELECTED;

	for(u32 i = 0; i < 3; i++) {
		TIMER[i].mode = TIMER_DISABLE;
		TIMER[i].max = 0xFFFF;
	}

	*COM_MODE = 2;
	*COM_DATA = 0xFF;
	*COM_CTRL1 = 0;
	*COM_CTRL2 = 3;

	*IOP_CTRL = 0x0F;
	*IOP_STOP = 0x62;
	*IOP_START = 0x0D;
	*DAC_CTRL = 0x00;
	*DAC_DATA = 0x00;
	*BATT_CTRL = 0x00;

	*IRDA_MODE = 0x0E;
	*IRDA_CTRL = 0x00;

	*CLK_MODE = CLKMODE_4MHZ;
	while(!(*CLK_MODE & CLKMODE_STABLE)) { /*nothing*/ }

	*F_WAIT2 = 0;
	*F_WAIT1 = 0;
	*F_CTRL = 0;
	*F_TAG = 0;
	for(u32 i = 0; i < 16; i++) {
		F_DATA[i] = 0x0F;
	}

	*LCD_CTRL = LCDCTRL_RR64HZ | LCDCTRL_ACTIVE | LCDCTRL_DMALL;
	*LCD_CAL = *F_CAL & 0x3F;

	for(u32 i = 0; i < 32; i++) {
		VRAM[i] = 0x00000000;
	}

	///// init_int_and_devs_first follows.

	*I_HCLR = 0x3FFF;
	// I_MCLR already set 3FFF.

	*I_MASK = INTMASK_IOP;

	// CustomHandlers[1] = 0; is already set(cleared)

	//svcChangeAutoDocking(KERNFLAGS_AUTODAC | KERNFLAGS_AUTOIR | KERNFLAGS_AUTOLCD);
	KernFlags = KERNFLAGS_AUTODAC | KERNFLAGS_AUTOIR | KERNFLAGS_AUTOLCD; // here, KernFlags == 0, so just setting value is ok.

	initDockRelatedDevices();

	// part of set_custom0swi_and_svcvector follows

	// CustomHandlers[0] = 0; is already set(cleared)
	// PtrToSvcVector = DefaultSvcVector; is done in main.S.
}
