#include "common_c.h"

extern u32 svcSetCpuSpeed(u32 speed);

void enableFlashWriteC() {
	*(volatile u16 *)0x080055AA = 0xFFAA;
	*(volatile u16 *)0x08002A54 = 0xFF55;
	*(volatile u16 *)0x080055AA = 0xFFA0;
}

void waitThenMakeFlashROC() {
	for(int i = 0; i < 800; i++) /*nothing*/ ;
	while(!(*F_WAIT2 & 0x08)) /*nothing*/ ;
	*F_WAIT2 = 0x00;
}

void setComOff() {
	*I_MCLR = INTMASK_COM;
	*I_HCLR = INTMASK_COM;

	*COM_MODE = 0x02;
	*COM_CTRL1 = 0x00;
	*COM_CTRL2 = 0x03;
	*COM_DATA = 0xFF;

	KernFlags &= ~KERNFLAGS_COMEN;
}

void setComOn() {
	*COM_CTRL1 = 0x00;
	*COM_MODE = 0x02;
	*COM_DATA = 0xFF;
	*COM_CTRL2 = 0x03;

	KernFlags &= ~KERNFLAGS_COMEN;

	byteF0 = 0x08;
	byteF1 = 0x00;

	u32 mask = *I_MASK;
	*I_MCLR = INTMASK_COM;
	*I_HCLR = INTMASK_COM;
	*I_MASK = mask | INTMASK_COM; // or???

	if(*I_STAT & INTMASK_IOP) {
		// if docked
		svcSetCpuSpeed(7);
		KernFlags |= KERNFLAGS_COMEN;
		*COM_MODE = 0x06;
		*COM_CTRL1 = 0x02;
		*I_MASK |= INTMASK_COM; // or???
	}
}

// @return -1:invalid 0:normal 1:snapshot
s32 getMCXType(u32 diridx) {
	// is its file exists and first block?
	if(((struct mcdir *)((u8 *)PFLA + 0x80 * diridx))->state != 0x51) {
		return -1;
	}
	const struct schdr *hdr = (void*)((u8 *)PFLA + 0x2000 * diridx);
	// is its file magic good?
	if(hdr->scmagic != 0x4353/*'SC'*/) {
		return -1;
	}
	// is its pk app magic good?
	// XXX differs from original.
	if((hdr->pkmagic & 0xFEFFffff) != 0x3058434D/*'MCX0'*/) {
		return -1;
	}
	// return its type.
	return (hdr->pkmagic & 0x01000000) ? 1 : 0;
}

// you can use this for getting offset to snapshot data
u32 getOffsetForSCData(u32 diridx) {
	u32 blkoff = diridx * 0x2000;
	u8 *pBlock = (u8*)PFLA + blkoff;
	u32 icons = ((struct schdr *)pBlock)->numicons & 0x0F;
	return blkoff + 0x80 + icons * 0x80;
}

u32 bcd2int2(u32 bcd) {
	return ((bcd >> 4) * 10) + (bcd & 0x0F);
}

u32 int2bcd2(u32 i) {
	// following expression costs a lot.
	//return ((i / 10) << 4) | (i % 10);
	u32 r = 0;
	while(10 <= i) {
		r += 0x10;
		i -= 10;
	}
	return r + i;
}

// original: decide leap by y2%4==0
//           not canonicalize bcd itself
u32 canonBCDDate_long(u32 bcddate) {
	u32 bcdret = 0;

	u32 bcdc = (bcddate >> 24) & 0xFF;
	u32 c = bcd2int2(bcdc);
	if(99 < c) {
		bcdret |= 0x99000000;
	} else {
		bcdret |= bcdc << 24;
	}

	u32 bcdy2 = (bcddate >> 16) & 0xFF;
	u32 y2 = bcd2int2(bcdy2);
	if(99 < y2) {
		bcdret |= 0x00990000;
	} else {
		bcdret |= bcdy2 << 16;
	}

	u32 bcdm = (bcddate >> 8) & 0xFF;
	u32 m = bcd2int2(bcdm);
	if(12 < m) {
		bcdret |= 0x00001200;
		m = 12;
	} else if(m < 1) {
		bcdret |= 0x00000100;
		m = 1;
	} else {
		bcdret |= bcdm << 8;
	}

	static const u8 maxdays[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	u8 maxd;
	if((m == 2) && (y2 % 4 == 0)) {
		maxd = 29;
	} else {
		maxd = maxdays[m];
	}

	u32 bcdd = (bcddate >> 0) & 0xFF;
	u32 d = bcd2int2(bcdd);
	if(maxd < d) {
		bcdd = int2bcd2(maxd);
	} else if(d < 1) {
		bcdd = 0x01;
	}
	bcdret |= bcdd;

	return bcdret;
}

static inline u32 clamp(u32 min, u32 val, u32 max) {
	return (val < min) ? min : (max < val) ? max : val;
}

#if 0
// requires umodsi3
static inline int isLeapYear(u32 year) {
	return (year % 400 == 0) || ((year % 100 != 0) && ((year % 4) == 0));
}

u32 strictCanonBCDDate(u32 bcddate) {
	u32 bcdret = 0;

	u32 c = clamp(0, bcd2int2((bcddate >> 24) & 0xFF), 99);
	bcdret |= int2bcd2(c) << 24;

	u32 y2 = clamp(0, bcd2int2((bcddate >> 16) & 0xFF), 99);
	bcdret |= int2bcd2(y2) << 16;

	u32 m = clamp(1, bcd2int2((bcddate >> 8) & 0xFF), 12);
	bcdret |= int2bcd2(m) << 8;

	static const u8 maxdays[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	u8 maxd;
	if(m == 2 && isLeapYear(c * 100 + y2)) {
		maxd = 29;
	} else {
		maxd = maxdays[m];
	}

	u32 d = clamp(1, bcd2int2((bcddate >> 0) & 0xFF), maxd);
	bcdret |= int2bcd2(d);

	return bcdret;
}

u32 canonBCDWeekTime_long(u32 bcdweektime) {
	u32 bcdret = 0;

	// week: clamp 1..7
	u32 bcdw = bcdweektime & 0xFF000000;
	if(bcdw < 0x01000000) {
		bcdret |= 0x01000000;
	} else if(0x07000000 < bcdw) {
		bcdret |= 0x07000000;
	} else {
		bcdret |= bcdw;
	}

	u32 bcdh = bcdweektime & 0x00FF0000;
	if(0x00230000 < bcdh) {
		bcdret |= 0x00230000;
	} else {
		bcdret |= bcdh;
	}

	u32 bcdm = bcdweektime & 0x0000FF00;
	if(0x00005900 < bcdm) {
		bcdret |= 0x00005900;
	} else {
		bcdret |= bcdm;
	}

	u32 bcds = bcdweektime & 0x000000FF;
	if(0x00000059 < bcds) {
		bcdret |= 0x00000059;
	} else {
		bcdret |= bcds;
	}

	return bcdret;
}
#endif

static inline u32 bcdClamp(u32 shift, u32 min, u32 value, u32 max) {
	return clamp(min << shift, value & (0xFF << shift), max << shift);
}

u32 canonBCDWeekTime(u32 bcdweektime) {
	u32 bcdret = 0;

	bcdret |= bcdClamp(24, 0x01, bcdweektime, 0x07); // week: 1..7
	bcdret |= bcdClamp(16, 0x00, bcdweektime, 0x23); // hour: 0..23
	bcdret |= bcdClamp( 8, 0x00, bcdweektime, 0x59); // minute: 0..59
	bcdret |= bcdClamp( 0, 0x00, bcdweektime, 0x59); // second: 0..59

	return bcdret;
}

u32 canonBCDDate(u32 bcddate) {
	u32 bcdret = 0;

	bcdret |= bcdClamp(24, 0x00, bcddate, 0x99); // century: 0..99
	bcdret |= bcdClamp(16, 0x00, bcddate, 0x99); // year2: 0..99
	bcdret |= bcdClamp( 8, 0x01, bcddate, 0x12); // month: 1..12

	static const u8 bcdmaxdays[12] = {0x31, 0x28, 0x31, 0x30, 0x31, 0x30, 0x31, 0x31, 0x30, 0x31, 0x30, 0x31};
	u8 bcdmaxd;
	u32 m = bcd2int2((bcdret >> 8) & 0xFF);
	if((m == 2) && ((bcd2int2((bcdret >> 16) & 0xFF) % 4) == 0)) { // XXX only %4 leap is checked here.
		bcdmaxd = 0x29;
	} else {
		bcdmaxd = bcdmaxdays[m - 1];
	}

	bcdret |= bcdClamp( 0, 0x01, bcddate, bcdmaxd); // day: 1..(depends by month)

	return bcdret;
}

