#include "common_S.h"

	// r12 is internal reg (we can destroy r12).
	// preserve r4-r11!

public_func svcSetCallbacks // index, ptr -> oldptr
	cmp r0, #3
	bhi panic

	.reloc ., R_ARM_ABS8, CustomHandlers
	mov r12, #0
	lsl r2, r0, #2
	ldr r0, [r12, r2] // return old ptr
	str r1, [r12, r2] // set new ptr
	bx lr

public_func svcCustomSwi2 // r0..r6,r8..r10 -> r0
	.reloc ., R_ARM_ABS8, CustomHandlers // +0
	mov r12, #0
	ldr r12, [r12]
	// lr is already good.
	bx r12

/*
public_func svcFlashWriteVirtual // sector, ptr -> ok?0:1 (equals to svcFlashWritePhysical)
	lsr r2, r0, #6 // sector2block (8192 / 128)
	ldr r3, =0x06000100 // F_DATA
	bx lr
*/

public_func svcSetCpuSpeed // speed -> oldspeed
	// change flash wait
	// if 0x08 <= speed? then F_WAIT1/2 = 0x10 else 0x00
	tst r0, #0x08
	moveq r1, #0x00
	movne r1, #0x10
	mov r2, #0x06000000
	str r1, [r2, #0x0C] // F_WAIT1
	str r1, [r2, #0x10] // F_WAIT2

	// if same clock then no-op.
	mov r1, #0x0B000000
	ldr r2, [r1] // CLK_MODE
	and r3, r2, #0x0F
	cmp r3, r0
	beq 1f

	// change clock
	str r0, [r1] // CLK_MODE

	// wait for stable clock
2:
	ldr r0, [r1] // CLK_MODE
	tst r0, #CLKMODE_STABLE
	beq 2b

	mov r0, r3 // return old CLK_MODE & 0x0F

1:
	bx lr

/*
public_func svcSenseAutoCom // void -> void
	bx lr
*/

public_func svcGetPtrToComFlags // void -> ptr (0xC0)
	.reloc ., R_ARM_ABS8, KernFlags
	mov r0, #0
	bx lr

public_func svcChangeAutoDocking // flags(16..18bit) -> flags & 0x00070000? void?
	mov r1, #0x00070000
	and r0, r1
	.reloc ., R_ARM_ABS8, KernFlags
	mov r2, #0
	ldr r3, [r2]
	bic r3, r1
	orr r3, r0
	str r3, [r2]
	bx lr

/*
public_func svcPrepareExecute // flag, appindex, param -> appindex (old or new)?
	bx lr
*/

public_func svcDoExecute // snapshot_save_flag -> void if COM-ing, otherwise noreturn.
	// if COM-ing enabled, executing is prohibited.
	.reloc ., R_ARM_ABS8, KernFlags
	mov r1, #0
	ldr r1, [r1]
	tst r1, #KERNFLAGS_COMEN
	bxne lr

	// disable FIQ
	mrs r1, cpsr
	orr r1, #PSR_FIQDIS
	msr cpsr_cf, r1

	// always make snapshot on SectBuf.
	// there are no other timing to snapshot registers...
	// it costs... but this is not hot code, it is ok...
	// that's "Premature Optimization".

	// we can't break r0 because it's an arg.
	ldm sp, {r1-r3}
	// we can break only r12 now... r0 is arg, r1-r3 restored, r4-r11,sp,lr are preserved.
	.reloc ., R_ARM_ABS8, SectBuf
	mov r12, #0
	stmia r12!, {r1-r11} // r1-r11 are saved; now we can break r1-r3

	ldr r1, [sp, #0x0C] // r12 after r1-r3
	stmia r12!, {r1} // r12 and user sp/lr
	// base writeback should not be used when accessing user bank; gas warns unpredictable behavior.
	stmia r12, {sp,lr}^ // r12 and user sp/lr
	add r12, #8

	ldr r1, [sp, #0x10] // lr(pc_usr) after r1-r3,r12
	mrs r2, spsr
	.reloc ., R_ARM_ABS8, SnapMagic
	mov r3, #0
	ldr r3, [r3]
	stmia r12!, {r1-r3}

	// ok all saved. go ahead.
	b doExecuteCont

public_func doExecute0Last // entrypoint
	// reset SV sp
	mov sp, #0x180

	// clearing user RAM
	mov r1, #0
	mov r2, #0x200
1:
	str r1, [r2], #4
	cmp r2, #0x0800
	blt 1b

	// umm, not enabled FIQ at boot on sv mode... is that ok?
	mrs r1, cpsr
	bic r1, #PSR_FIQDIS
	msr cpsr_cf, r1

	// escape r0(entrypoint) to r2 to use r0 as param for the app.
	mov r2, r0

	// load param for the app from PreparedAppParam
	.reloc ., R_ARM_ABS8, PreparedAppParam
	mov r0, #0
	ldr r0, [r0]

	// go to user mode
	mov r1, #PSR_USER
	msr cpsr_cf, r1

	// reset USR sp
	mov sp, #0x0800

/*
stmfd sp!, {r0-lr}
mov r0, #0x02000000
ldr r1, =0x0D000100
mov r2, #32
1:
	ldr r3, [r0, #4]!
	str r3, [r1, #4]!
	subs r2, #1
	bne 1b
b .
.pool
//*/

	// GO!
	bx r2

public_func doExecute1Last // entrypoint
	// XXX original does not reset SV sp!
	// mov sp, #0x180

	// umm, not disabled FIQ at boot... is that ok?
	mrs r0, cpsr
	bic r0, #PSR_FIQDIS
	msr cpsr_cf, r0

	// restore regs from SectBuf; here, we can break almost all regs
	.reloc ., R_ARM_ABS8, SectBuf
	mov r0, #0x3C

	ldr r1, [r0]
	msr spsr_cf, r1

	ldmdb r0!, {lr}

	// base writeback should not be used when accessing user bank; gas warns unpredictable behabior.
	ldmdb r0, {r1-r12, sp, lr}^ // note: do not touch banked regs after this (LDM that restores user bank regs)
	nop

	movs pc, lr // jump with psr resume

public_func svcFlashReadSerial // void -> serial
	// original krom reads F_SN as two halfword... something hardware limitation? 16bit data bus?
	// ldr can specify 12bit offset but ldrh can only 8bit.
#if 0
	mov r1, #0x06000000 // F_SN
	ldr r0, [r1, #0x300]
#else
	ldr r1, =0x06000300 // F_SN
	ldrh r0, [r1, #2]
	lsl r0, #16
	ldrh r1, [r1, #0]
	orr r0, r1
#endif
	bx lr

	.pool

public_func svcAllowMCWriteAltSector // void -> new comflags with bit10=0
	.reloc ., R_ARM_ABS8, KernFlags
	mov r1, #0
	ldr r0, [r1]
	bic r0, #KERNFLAGS_WPALT
	str r0, [r1]
	bx lr

/*
public_func svcSetBcdDateTime // date, time -> void
	bx lr
*/

public_func svcGetBcdDate // void -> date
	mov r0, #0x0B800000

	// get stable date? XXX timing?
1:
	ldr r1, [r0, #0x0C] // RTC_DATE
	ldr r2, [r0, #0x0C] // RTC_DATE
	cmp r1, r2
	bne 1b

	bic r1, #0xFF000000
	.reloc ., R_ARM_ABS8, CurrentCenturyBCD
	mov r0, #0
	ldrb r0, [r0]
	orr r0, r1, r0, lsl #24

	bx lr

public_func svcGetBcdTime // void -> time
	mov r2, #0x0B800000

	// get stable time? XXX timing?
1:
	ldr r0, [r2, #0x08] // RTC_TIME
	ldr r1, [r2, #0x08] // RTC_TIME
	cmp r0, r1
	bne 1b

	bx lr

public_func svcFlashWriteSerial // serial -> 0? void?
	push {lr}

	mov r3, #0x06000000
	ldr r1, [r3, #0x308] // F_CAL
	and r1, #0x3F

	mov r2, #0x41
	str r2, [r3, #0x10] // F_WAIT2/control?

	bl enableFlashWrite

	mov r2, #0x08000000 // PFLA
	// original krom writes F_SN as two haldword... something hardware limitation? 16bit data bus?
#if 1
	strh r0, [r2]
	lsr r0, #16
	strh r0, [r2, #2]
#else
	str r0, [r2]
#endif
	// and F_CAL.
	strh r1, [r2, #8]

	pop {lr}
	b waitThenMakeFlashRO

// note: original doesn't handle bad sectors.
public_func svcFlashWritePhysical // sector, ptr -> ok?0:1
	push {lr}

	mov r2, #0x21
	mov r3, #0x06000000
	str r2, [r3, #0x10] // F_WAIT2

	bl enableFlashWrite

	// transfer data
	lsl r0, #7 // log2(2048/128)
	add r0, #0x08000000 // PFLA
	// original krom writes data as halfwords... something hardware limitation? 16bit data bus?
#if 1
	mov r2, #0x80/2
1:
	ldrh r3, [r1], #2
	strh r3, [r0], #2
	subs r2, #1
	bne 1b
#else
	push {r4-r11}

	mov r2, #0x80/32
1:
	ldmia r1!, {r4-r11}
	stmia r0!, {r4-r11}
	subs r2, #1
	bne 1b

	pop {r4-r11}
#endif

	bl waitThenMakeFlashRO

	// verify data
	mov r2, #0x80/4
1:
	// TODO is it safe to verify backward?
	ldr r3, [r0, #-4]!
	ldr lr, [r1, #-4]!
	cmp r3, lr
	bne 2f // verify error!

	subs r2, #1
	bpl 1b

	// ok
	mov r0, #0
	pop {pc}

	// verify error
2:
	mov r0, #1
	pop {pc}

public_func svcSetComOnOff // flag -> void?
	cmp r0, #0
	// tailcall
	beq setComOff
	b setComOn

/*
public_func svcTestSnapshot // appindex -> normal?0:good?1:?
	//TODO
	bx lr
*/

public_func svcGetPtrToAlarmSetting // void -> ptr (0xD8)
	.reloc ., R_ARM_ABS8, AlarmSettings
	mov r0, #0
	bx lr

public_func svcGetPtrToPtrToSwiTable // void -> ptr (0xE0)
	.reloc ., R_ARM_ABS8, PtrToSvcVector
	mov r0, #0
	bx lr

/*
public_func svcMakeAlternateDirIndex // do_set_flag, appidx -> appidx (new or old)?
	.reloc ., R_ARM_ABS8, altAppIndex
	mov r2, #0

	cmp r0, #1
	ldrne r0, [r2]
	bxne lr

	cmp r1, #0
	streq r1, [r2]
	moveq r0, r1

	bx lr
*/

public_func svcGetDirIndex // void -> appindex (or alternate?)
	.reloc ., R_ARM_ABS8, CurrentExecAppIndex
	mov r0, #0
	ldrh r0, [r0]
	cmp r0, #0
	bxne lr

	.reloc ., R_ARM_ABS8, AltAppIndex
	mov r0, #0
	ldrb r0, [r0]
	bx lr

public_func svcGetPtrToFunc3addr // void -> ptr (0xC8)
	.reloc ., R_ARM_ABS8, BUFunc3offset
	mov r0, #0
	bx lr

public_func svcFlashReadWhateverByte // sector -> ((b8*)PFLA)[sector*0x80+0x7E]
	lsl r0, #7
	add r0, #0x08000000 // PFLA
	ldrh r0, [r0, #0x7E] // Flash must be accessed by halfword?
	and r0, #0xFF
	bx lr

public_rodata DefaultSvcVector
	.word 0                        // 00: invalid (reset)
	.word svcSetCallbacks          // 01: set custom handler
	.word svcCustomSwi2            // 02: call svchandler
	.word svcFlashWriteVirtual     // 03: write a sector to flash (virtual address)
	.word svcSetCpuSpeed           // 04: set CPU clock
	.word svcSenseAutoCom          // 05: TODO senseAutoCom
	.word svcGetPtrToComFlags      // 06: get ptr to kernel flags(C0)
	.word svcChangeAutoDocking     // 07: set kernel flags with param & 0007`0000 TODO ChangeAutoDocking?
	.word svcPrepareExecute        // 08: prepareExecute
	.word svcDoExecute             // 09: doExecute
	.word svcFlashReadSerial       // 0A: read flash serial
	.word svcAllowMCWriteAltSector // 0B: clears KERNFLAGS_WPALT: clear kernel flags bit 10
	.word svcSetBcdDateTime        // 0C: set date and time
	.word svcGetBcdDate            // 0D: get date
	.word svcGetBcdTime            // 0E: get time
	.word svcFlashWriteSerial      // 0F: write flash serial (deprecated)
	.word svcFlashWritePhysical    // 10: write a sector to flash (physical address)
	.word svcSetComOnOff           // 11: TODO setComOnOff
	.word svcTestSnapshot          // 12: test snapshot
	.word svcGetPtrToAlarmSetting  // 13: get ptr to alarm setting
	.word svcGetPtrToPtrToSwiTable // 14: get ptr to "ptr to svc handlers"
	.word svcMakeAlternateDirIndex // 15: TODO makeAlternateDirIndex
	.word svcGetDirIndex           // 16: TODO getDirIndex
	.word svcGetPtrToFunc3addr     // 17: TODO getPtrToFunc3Addr
	.word svcFlashReadWhateverByte // 18: read last haldbyte of flash sector
	//.set sizeOfDefaultSvcVector, (. - defaultSvcVector) / 4
