#include "common_c.h"

extern u32 svcFlashWritePhysical(u32 sector, void *ptr);
extern s32 svcSetCpuSpeed(u32 speed);
extern u32 svcGetDirIndex();
extern void doExecuteSnapshotToSectBuf(void *enter_sp);
extern void __attribute((noreturn)) doExecute0Last(u32 entrypoint);
extern void __attribute((noreturn)) doExecute1Last();
extern void setComOff();
extern void setComOn();
extern s32 getMCXType(u32 diridx);
extern s32 getOffsetForSCData(u32 diridx);
extern u32 canonBCDDate(u32 bcddate);
extern u32 canonBCDWeekTime(u32 bcdweektime);
extern void clearCustomHandlers();

u32 svcFlashWriteVirtual(u32 sector, void *ptr) {
	u32 block = sector * 128 / 8192; // v_block
	for(u32 i = 0; ; i++) {
		u32 v = F_DATA[i];
		if(v == block) {
			// p_block i have the target sector.
			block = i; // v_block -> p_block
			break;
		}
		if(v == i + 1) {
			// ???
			return 1;
		}
		if(15 < i) {
			panic();
		}
	}

	if(!(*F_TAG & (1 << block))) {
		// this map is invalid...
		return 1;
	}

	return svcFlashWritePhysical(block * 8192 / 128 + sector % (8192 / 128), ptr);
}

void svcSenseAutoCom() {
	if(!(KernFlags & KERNFLAGS_SENSERQ)) {
		return;
	}

	KernFlags &= ~KERNFLAGS_SENSERQ;
	if(*I_STAT & INTMASK_COM) {
		BUFunc3offset = 0;
		//byteCA = 0; // byteCA is never read; skipping write.
	}
	setComOn();
}

// @param flags only bit16..18(0x0007`0000) are touched
// @return maybe void?
u32 svcChangeAutoDockingC(u32 flags) {
	KernFlags = (KernFlags & ~0x00070000) | (flags & 0x00070000);
	return (flags & 0x00070000);
}

u32 svcPrepareExecute(u32 do_set_diridx, u16 diridx, u32 param) {
	if(do_set_diridx == 1) {
		if(diridx != 0) {
			if(getMCXType(diridx) == -1) {
				// invalid; abort setting
				return PreparedAppIndex;
			}
		}
		PreparedAppIndex = diridx;
		PreparedAppParam = param;
	}
	return PreparedAppIndex;
}

void doExecuteCont(u32 snapshot_save_flag) {
	setComOff();

	*I_MASK = INTMASK_RTC | INTMASK_IOP;
	*I_MCLR = ~(INTMASK_RTC | INTMASK_IOP);
	*I_HCLR = ~(INTMASK_RTC | INTMASK_IOP);

	svcSetCpuSpeed(7);

	*F_WAIT1 = 0x00;
	*F_WAIT2 = 0x00;

	clearCustomHandlers();

	// SE\0\0 or SE\0\001
	SnapMagic = (SnapMagic & 0xFFFFfffe) | (snapshot_save_flag ? 1 : 0);

	KernFlags &= ~(KERNFLAGS_PREPARED | KERNFLAGS_MCRW);

	// save snapahot if current app is not GUI and apptype MCX1.
	u8 curapp = svcGetDirIndex();
	if(curapp != 0 && getMCXType(curapp) == 1) {
		u32 dstsect = getOffsetForSCData(curapp) / 0x80;
		// save r1-12,sp_usr,lr_usr,lr_sv(pc_usr),spsr,SnapMagic
		// already snapshotted at SectBuf at entry of svcDoExecute.
		svcFlashWritePhysical(dstsect, SectBuf); // XXX no result check!
		dstsect += 4; // yes, it's 4 sectors.
		// save user mem (200-7ff)
		for(u8 *p = (u8*)0x200; p < (u8*)0x800; p += 0x80) {
			svcFlashWritePhysical(dstsect, p); // XXX no result check!
			dstsect++;
		}
	}

	// start executing new application

	CurrentExecAppIndex = PreparedAppIndex;
	PreparedAppIndex = 0;

	*F_STAT; // dummy read
	*F_CTRL = 1;
	*F_TAG = 0;
	for(u32 i = 0; i < 16; i++) {
		F_DATA[i] = 0x0F;
	}

	curapp = svcGetDirIndex(); // considering altappindex

	u32 entrypoint;

	if(curapp == 0) {
		entrypoint = 0x04001E00;
	} else {
		// map flash blocks of current app as contiguous.
		u32 block = CurrentExecAppIndex;
		u32 tag = 0;
		for(u32 i = 0; ; i++) {
			F_DATA[block] = i;
			tag |= 1 << block;
			block = ((struct mcdir *)((u8*)PFLA + 0x80 * block))->nextblock;
			if(block == 0xFFFF) {
				break;
			}
			block++; // note: nextblock is 0 origin
		}
		*F_TAG = tag;
		*F_CTRL = 2;

		entrypoint = ((struct schdr *)((u8*)PFLA + 0x2000 * curapp))->entrypoint;

		if(getMCXType(curapp) == 1) {
			u32 dataoff = getOffsetForSCData(curapp);
			u8 *pdata = (u8*)PFLA + dataoff;
			if(*(u32*)(pdata + 0x40) == 0x45530001/*'ES\0\1' "\1\0SE"*/) {
				// load snapshot

				// copy reg-area to buffer for rewriting the reg-area sector.
				for(u32 i = 0; i < 0x80; i += 4) {
					*(u32*)(SectBuf + i) = *(u32*)(pdata + i);
				}
				*(u32*)(SectBuf + 0x40) &= ~0x00000001; // ES01->ES00
				svcFlashWritePhysical(dataoff / 0x80, SectBuf); // XXX no result check!

				// restore user ram XXX original transfers snapshot+0 -> 200~7FF, not from snapshot+200!
				for(u32 i = 0x200; i < 0x800; i += 4) {
					*(u32*)(0x200 + i) = *(u32*)(pdata + i - 0x200/*as is, but it is wrong offset!*/);
				}

				SnapMagic = *(u32*)(pdata + 0x40); // re-read from Flash?
				doExecute1Last(entrypoint);
			}
			// fall-through to MCX0
		}
	}

	// GUI or MCX0 (or even invalid app? no. svcPrepareExecute refuses that case.)
	SnapMagic &= 0xFFFFfffe; // clear bit0
	doExecute0Last(entrypoint);
}

static inline void waitRTC() {
	while(!(*I_STAT & INTMASK_RTC)) { /*nothing*/ }
}

static inline void fwdRTC(u32 mode, volatile u32 *pfield, u32 shift, u32 value) {
	*RTC_MODE = mode;

	for(;;) {
		waitRTC();
		if((*pfield & (0xFF << shift)) == (value & (0xFF << shift))) {
			break;
		}
		// XXX original: twice for what?
		*RTC_CTRL = 1;
		*RTC_CTRL = 1;
	}
}

void svcSetBcdDateTime(u32 bcd_date, u32 bcd_weektime) {
	bcd_date = canonBCDDate(bcd_date);
	bcd_weektime = canonBCDWeekTime(bcd_weektime);

	CurrentCenturyBCD = bcd_date >> 24;

	fwdRTC(RTCMODE_PAUSE | RTCMODE_DAY, RTC_DATE,  0, 1);
	fwdRTC(RTCMODE_PAUSE | RTCMODE_YEAR, RTC_DATE, 16, bcd_date);
	fwdRTC(RTCMODE_PAUSE | RTCMODE_MONTH, RTC_DATE,  8, bcd_date);
	fwdRTC(RTCMODE_PAUSE | RTCMODE_DAY, RTC_DATE,  0, bcd_date);
	fwdRTC(RTCMODE_PAUSE | RTCMODE_DOW, RTC_TIME, 24, bcd_weektime);
	fwdRTC(RTCMODE_PAUSE | RTCMODE_HOUR, RTC_TIME, 16, bcd_weektime);
	fwdRTC(RTCMODE_PAUSE | RTCMODE_MINUTE, RTC_TIME,  8, bcd_weektime);
	fwdRTC(RTCMODE_PAUSE | RTCMODE_SECOND, RTC_TIME,  0, bcd_weektime);
	// wait for 0.5-1.0sec
	waitRTC();
	while(*I_STAT & INTMASK_RTC) { /*nothing*/ } // almost same as waitRTC but condition is inverted.

	*RTC_MODE = RTCMODE_RUN | RTCMODE_UNSELECTED;
	LastKnownYear2BCD = (*RTC_DATE >> 0x10) & 0xFF;
}

// @return: 1 if it is MCX1 and have valid snapshot magic ("ES\0\1"); 0 otherwise.
u32 svcTestSnapshot(u32 dirindex) {
	if(getMCXType(dirindex) != 1) {
		return 0;
	}

	u32 dataoff = getOffsetForSCData(dirindex);
	u8 *pdata = (u8*)PFLA + dataoff;
	if(*(u32*)(pdata + 0x40) != 0x45530001/*'\1\0SE' "ES\0\1"*/) {
		return 0;
	}

	return 1;
}

u32 svcMakeAlternateDirIndex(u32 do_set_appidx, u32 appidx) {
	if(do_set_appidx == 1) {
		if(appidx != 0) {
			if(svcGetDirIndex() != appidx) {
				return AltAppIndex;
			}
		}
		AltAppIndex = appidx;
	}
	return AltAppIndex;
}

u8 svcFlashReadWhateverByteC(u32 sector) {
	// without volatile, it access by byte.
	// PFLA requires halfword access? really?? other routines use byte/word...
	u16 r = ((const volatile u16 *)PFLA)[(sector * 128 + 0x7E) / 2];
	return r & 0xFF;
}
