// asm & C shared header

///// CPU specific

#define PSR_USER  0x10
#define PSR_FIQ   0x11
#define PSR_IRQ   0x12
#define PSR_SV    0x13
#define PSR_ABORT 0x17
#define PSR_UNDEF 0x1B
#define PSR_SYS   0x1F

#define PSR_THUMB 0x20

#define PSR_FIQDIS 0x40
#define PSR_IRQDIS 0x80

#define PSR_V 0x10000000
#define PSR_C 0x20000000
#define PSR_Z 0x40000000
#define PSR_N 0x80000000

///// Hardware specific

// TODO
#define FSTAT_
#define FCTRL_

#define INTMASK_ACTION 0x0001
#define INTMASK_RIGHT  0x0002
#define INTMASK_LEFT   0x0004
#define INTMASK_DOWN   0x0008
#define INTMASK_UP     0x0010
// 0x0020 Unknown
#define INTMASK_COM    0x0040 // MC Com (/SEL pin?) // FIQ
#define INTMASK_TIMER0 0x0080
#define INTMASK_TIMER1 0x0100
#define INTMASK_RTC    0x0200
#define INTMASK_BATTLO 0x0400
#define INTMASK_IOP    0x0800 // Docked (VCC pin? IO Power?)
#define INTMASK_IRRX   0x1000
#define INTMASK_TIMER2 0x2000 // FIQ

#define TIMER_DIV2 0
#define TIMER_DIV32 1
#define TIMER_DIV512 2
#define TIMER_ENABLE 0x04
#define TIMER_DISABLE 0x00 // bit2 off

#define CLKMODE_64KHZ 1
#define CLKMODE_128KHZ 2
#define CLKMODE_256KHZ 3
#define CLKMODE_512KHZ 4
#define CLKMODE_1MHZ 5
#define CLKMODE_2MHZ 6
#define CLKMODE_4MHZ 7
#define CLKMODE_8MHZ 8 // TODO is this able to use??
#define CLKMODE_STABLE 0x10

#define RTCMODE_RUN        0 // bit0 off
#define RTCMODE_PAUSE      1
#define RTCMODE_SECOND     (0 << 1) // bit1-3 off
#define RTCMODE_MINUTE     (1 << 1)
#define RTCMODE_HOUR       (2 << 1)
#define RTCMODE_DOW        (3 << 1)
#define RTCMODE_DAY        (4 << 1)
#define RTCMODE_MONTH      (5 << 1)
#define RTCMODE_YEAR       (6 << 1)
#define RTCMODE_UNSELECTED (7 << 1)

#define COMMODE_DATA 0x01
#define COMMODE_ACK  0x02
// TODO
#define COMCTRL1_
#define COMCTRL2_

#define IRDAMODE_DIRRX 0x00 // ~0x01
#define IRDAMODE_DIRTX 0x01
#define IRDAMODE_DISABLE 0x02

#define LCDCTRL_DMALL  0x00 // (bit0-2 = 0) all 32 rows on (NO$)
#define LCDCTRL_DM1_8  0x01 // (bit0-2 = 1) first 8 rows on (NO$)
#define LCDCTRL_DM2_8  0x02 // (bit0-2 = 2) second 8 rows on (NO$)
#define LCDCTRL_DM3_8  0x03 // (bit0-2 = 3) third 8 rows on (NO$)
#define LCDCTRL_DM4_8  0x04 // (bit0-2 = 4) fourth 8 rows on (NO$)
#define LCDCTRL_DMT_16 0x05 // (bit0-2 = 5) first 16 rows on (NO$)
#define LCDCTRL_DMM_16 0x06 // (bit0-2 = 6) middle 16 rows on (NO$)
#define LCDCTRL_DMB_16 0x07 // (bit0-2 = 7) bottom 16 rows on (NO$)
#define LCDCTRL_ENABLE 0x08 // if 0, weird fadeout (NO$)
#define LCDCTRL_RR64HZ 0x10 // (bit4-5 = 1)
#define LCDCTRL_RR32HZ 0x20 // (bit4-5 = 2)
#define LCDCTRL_RR16HZ 0x30 // (bit4-5 = 3)
#define LCDCTRL_ACTIVE 0x40 // if 0, no dots are drawn.
#define LCDCTRL_ROTATE 0x80

#define IOPPIN_REDLED 0x02
#define IOPPIN_SOUND  0x20

#define DACCTRL_DISABLE 0x00 // bit0 off
#define DACCTRL_ENABLE  0x01

///// Kernel specific

#define KERNFLAGS_0001     0x00000001 // ? comm5A,5F(not 5E)
#define KERNFLAGS_0002     0x00000002 // ? comm5A,5E
#define KERNFLAGS_0004     0x00000004 // ? comm5A,5E
#define KERNFLAGS_0008     0x00000008 // ? comm5A,5E
#define KERNFLAGS_SENSERQ  0x00000100 // If true, svcSenseAutoCom does some work.
#define KERNFLAGS_COMEN    0x00000200 // COM (MC interface) enable
#define KERNFLAGS_WPALT    0x00000400 // MC Write protect for alt-sector area? (not svcFlashWritePhys)
#define KERNFLAGS_PREPARED 0x00000800 // App is prepared by BU command 59
#define KERNFLAGS_AUTODAC  0x00010000 // Auto stop DAC when dock/undock
#define KERNFLAGS_AUTOIR   0x00020000 // Auto stop IR when dock/undock
#define KERNFLAGS_AUTOLCD  0x00040000 // Auto rotate LCD when dock/undock
#define KERNFLAGS_MCRW     0x10000000 // Has MemoryCard R/W(/S) (formerly Downloading pk app)
#define KERNFLAGS_SETRTC   0x20000000 // Set RTC from SectBuf after MC transaction
#define KERNFLAGS_REBOOT   0x40000000 // Restart to GUI after COM before call user FIQ handler

// note: BUFUNC_TOPS/BUFUNC_FROMPS are exclusive. BUFUNC_STAGE1/BUFUNC_STAGE2 are also exclusive.
#define BUFUNC_TOPS   0x01
#define BUFUNC_FROMPS 0x02
#define BUFUNC_ERROR  0x04
#define BUFUNC_STAGE1 0x08
#define BUFUNC_STAGE2 0x10
