#include <stdint.h>
#include "krom.h"

#define between(a,v,b) (((a) <= (v)) && ((v) <= (b)))

typedef uint8_t u8;
typedef uint16_t u16;
typedef uint32_t u32;
typedef uint64_t u64;
typedef int8_t s8;
typedef int16_t s16;
typedef int32_t s32;
typedef int64_t s64;

typedef struct bufunc1ret *(BUFunc)(u32 flags, void* buffer);

struct __attribute__((packed, aligned(4))) mcdir {
	u32 state;
	u32 size;
	u16 nextblock; // note: 0 origin
	u8 name[21]; // null terminated
	u8 zero;
	u8 reserved[0x5F];
	u8 checksum; // xor from top to just before this
};

struct __attribute__((packed, aligned(4))) schdr {
	u16 scmagic;
	u8  numicons;
	u8  numblocks;
	u8  title[64];
	u8  reserved_sc[12];
	u16 numfvicons;
	u32 pkmagic; // MCX0(pk-app), MCX1(pk-snapshot) or CRD0(pk-icon)
	u8  numexicons;
	u8  numbucommands;
	u32 reserved_pk;
	u32 entrypoint;
	/* SC palette 0x20 */
	/* SC icons 0x80*(numicons&0x0F) */
	/* PK snapshot 0x0800*(pkmagic=="MCX1"?1:0) */
	/* PK bucommands 8*ceil(numbucommands,0x10) */
	/* PK fvicon images 0x80*numfvicons */
	/* PK exicon lists 8*numexicons */
};

struct __attribute__((packed, aligned(4))) exiconentry {
	u16 animLength;
	u16 animDelay; // * 1/30 sec
	const u32 *addr;
};

struct __attribute__((packed, aligned(4))) bufuncentry {
	u32 paramLen;
	BUFunc *pFunc;
};

struct __attribute__((packed, aligned(4))) bufunc1ret {
	void *pBuf;
	u32 bufSize;
};

struct __attribute__((packed, aligned(4))) alarm_info {
	u8 minuteBCD;
	u8 hourBCD;

	u8 enabled : 1;
	u8 keylock : 1;
	u8 attenate : 2; // 0:loud 1:med 2:mute
	u8 : 3; // unknown
	u8 goodrtc : 1; // datetime has set

	u8 : 8; // unknown
	u32 offfont; // when use, bit-or 0x04000000(KROM).
};

struct __attribute__((packed, aligned(4))) reg_timer {
	u32 max;
	const u32 current;
	u32 mode;
	const u32 _padding;
};

static          const u16 * const VFLA   = (void*)0x02000000; // TODO writable?

// 0x04000000+0x020000: KROM

static volatile       u32 * const F_CTRL  = (void*)0x06000000;
static volatile       u32 * const F_STAT  = (void*)0x06000004;
static volatile       u32 * const F_TAG   = (void*)0x06000008;
static volatile       u32 * const F_WAIT1 = (void*)0x0600000C;
static volatile       u32 * const F_WAIT2 = (void*)0x06000010; // TODO wait? ctrl2?

// F_DATA[p_block] = v_block
static volatile       u32 * const F_DATA = (void*)0x06000100;

static volatile       u32 * const F_SN    = (void*)0x06000300;
static volatile       u32 * const F_CAL   = (void*)0x06000308;

static                u16 * const PFLA   = (void*)0x08000000;

static volatile const u32 * const I_HOLD = (void*)0x0A000000;
static volatile const u32 * const I_STAT = (void*)0x0A000004;
static volatile       u32 * const I_MASK = (void*)0x0A000008;
static volatile       u32 * const I_MCLR = (void*)0x0A00000C;
static volatile       u32 * const I_HCLR = (void*)0x0A000010;

static volatile       struct reg_timer * const TIMER = (void*)0x0A800000;

static volatile       u32 * const CLK_MODE = (void*)0x0B000000;

static volatile       u32 * const RTC_MODE = (void*)0x0B800000;
static volatile       u32 * const RTC_CTRL = (void*)0x0B800004;
static volatile       u32 * const RTC_TIME = (void*)0x0B800008;
static volatile       u32 * const RTC_DATE = (void*)0x0B80000C;

static volatile       u32 * const COM_MODE  = (void*)0x0C000000;
static volatile const u32 * const COM_STAT1 = (void*)0x0C000004;
static volatile       u32 * const COM_DATA  = (void*)0x0C000008;
static volatile       u32 * const COM_CTRL1 = (void*)0x0C000010;
static volatile const u32 * const COM_STAT2 = (void*)0x0C000014;
static volatile       u32 * const COM_CTRL2 = (void*)0x0C000018;

static volatile       u32 * const IRDA_MODE = (void*)0x0C800000;
static volatile       u32 * const IRDA_CTRL = (void*)0x0C800004;

static volatile       u32 * const LCD_CTRL = (void*)0x0D000000;
static volatile       u32 * const LCD_CAL  = (void*)0x0D000004;

static volatile       u32 * const VRAM     = (void*)0x0D000100;

static volatile       u32 * const IOP_CTRL  = (void*)0x0D800000;
static volatile       u32 * const IOP_STOP  = (void*)0x0D800004;
static volatile       u32 * const IOP_START = (void*)0x0D800008;
static volatile       u32 * const IOP_DATA  = (void*)0x0D80000C; // ?
static volatile       u32 * const DAC_CTRL  = (void*)0x0D800010;
static volatile       u32 * const DAC_DATA  = (void*)0x0D800014;
static volatile       u32 * const BATT_CTRL = (void*)0x0D800020;
