#include <inttypes.h>
#include "images.c"

#define min(x, y) ((x) < (y) ? (x) : (y))
#define max(x, y) ((x) > (y) ? (x) : (y))

#define TIMER_DIV2 0
#define TIMER_DIV32 1
#define TIMER_DIV512 2

#define CPUCLOCK_64KHZ 1
#define CPUCLOCK_128KHZ 2
#define CPUCLOCK_256KHZ 3
#define CPUCLOCK_512KHZ 4
#define CPUCLOCK_1MHZ 5
#define CPUCLOCK_2MHZ 6
#define CPUCLOCK_4MHZ 7

#define HANDLER_SWI 0
#define HANDLER_IRQ 1
#define HANDLER_FIQ 2
#define HANDLER_DOWNLOAD 3

#define IRQ_ACTION  0x0001
#define IRQ_RIGHT   0x0002
#define IRQ_LEFT    0x0004
#define IRQ_DOWN    0x0008
#define IRQ_UP      0x0010
//                  0x0020
#define IRQ_COMM    0x0040
#define IRQ_TIMER0  0x0080
#define IRQ_TIMER1  0x0100
#define IRQ_RTC     0x0200
#define IRQ_BATTLOW 0x0400
#define IRQ_IOP     0x0800
#define IRQ_IRDA    0x1000
#define IRQ_TIMER2  0x2000

#define LCDCTRL_DMALL  0x00 // (bit0-2 = 0) all 32 rows on (NO$)
#define LCDCTRL_DM1_8  0x01 // (bit0-2 = 1) first 8 rows on (NO$)
#define LCDCTRL_DM2_8  0x02 // (bit0-2 = 2) second 8 rows on (NO$)
#define LCDCTRL_DM3_8  0x03 // (bit0-2 = 3) third 8 rows on (NO$)
#define LCDCTRL_DM4_8  0x04 // (bit0-2 = 4) fourth 8 rows on (NO$)
#define LCDCTRL_DMT_16 0x05 // (bit0-2 = 5) first 16 rows on (NO$)
#define LCDCTRL_DMM_16 0x06 // (bit0-2 = 6) middle 16 rows on (NO$)
#define LCDCTRL_DMB_16 0x07 // (bit0-2 = 7) bottom 16 rows on (NO$)
#define LCDCTRL_ENABLE 0x08 // if 0, weird fadeout (NO$)
#define LCDCTRL_RR64HZ 0x10 // (bit4-5 = 1)
#define LCDCTRL_RR32HZ 0x20 // (bit4-5 = 2)
#define LCDCTRL_RR16HZ 0x30 // (bit4-5 = 3)
#define LCDCTRL_ACTIVE 0x40 // if 0, no dots are drawn.
#define LCDCTRL_ROTATE 0x80

#define IOPPIN_SOUND 0x20
#define DACCTRL_DISABLE 0x00 // bit0 clear
#define DACCTRL_ENABLE 0x01

#define CARD_NONE 0
#define CARD_HIME 1
#define CARD_BOZU 2
#define CARD_HAT 3
#define CARD_ARABIC 4

#define INLINE static inline

uint8_t g_attenate;
volatile uint32_t g_latch; // with sync = MSbit
volatile struct adpcm {
	const uint8_t *psound, *psoundend;
	uint8_t buf;
	int32_t predict, presample;
} g_adpcm;
uint32_t g_rand_val;
uint32_t g_offscreen[32];

uint8_t g_cards[100];
uint8_t g_lcards, g_rcards;
uint8_t g_mycards, g_cpucards, g_potcards;
uint8_t g_gameside, g_cpuside;
uint32_t g_frame; // used by each state

const uint32_t numfont[10] = {
	0x00075557,
	0x00044444,
	0x00071747,
	0x00074747,
	0x00044755,
	0x00074717,
	0x00075717,
	0x00044457,
	0x00075757,
	0x00074757,
};

extern void (*low_handler_ptrs[4])(void);

#define defsnd(name) extern uint8_t name##_start[], name##_end[]
defsnd(iyo4000_adpcm);
defsnd(ponr4000_adpcm);
defsnd(deden6000_adpcm);
defsnd(foshu4000_adpcm);
defsnd(fan6000_adpcm);
defsnd(mecha4000_adpcm);
defsnd(panchi4000_adpcm);

struct c_image_opaque {
	const uint32_t width, height;
	const uint32_t * const bit;
};
struct c_image_transp {
	struct c_image_opaque opq;
	const uint32_t * const rmask;
};
#define defimg_o(name) //const struct c_image_opaque name = { name##_width, name##_height, name##_bit }
#define defimg_t(name) //const struct c_image_transp name = { { name##_width, name##_height, name##_bit }, name##_rmask }
#define image_ctext image_10
defimg_o(image_10);
#define image_cbozu image_11
defimg_o(image_11);
#define image_chat image_12
defimg_o(image_12);
#define image_ctenno image_13
defimg_o(image_13);
#define image_chime image_14
defimg_o(image_14);
#define image_hime image_21
defimg_t(image_21);
#define image_bo image_22
defimg_t(image_22);
#define image_zu image_23
defimg_t(image_23);
#define image_win image_24
defimg_o(image_24);
#define image_lose image_25
defimg_o(image_25);
#define image_draw image_26
defimg_o(image_26);
#define image_hand1 image_31
defimg_t(image_31);
#define image_hand2 image_32
defimg_t(image_32);
#define image_hand3 image_33
defimg_t(image_33);
#define image_hand4 image_34
defimg_t(image_34);
#define image_hand5 image_35
defimg_o(image_35);
#define image_hand6 image_36
defimg_o(image_36);
#define image_cpucatch image_40
defimg_t(image_40);
#define image_cpufree image_41
defimg_t(image_41);
#define image_cleft image_50
defimg_t(image_50);
#define image_cright image_51
defimg_t(image_51);
#define image_carrow image_52
defimg_o(image_52);

volatile struct reg_intr {
	uint32_t latch;
	uint32_t raw;
	union {
		uint32_t mask; // for read
		uint32_t enable; // for write
	};
	uint32_t disable;
	uint32_t ack;
} * const reg_intr = (struct reg_intr *)0x0A000000;

volatile struct reg_lcd {
	uint32_t ctrl;
	uint32_t calibrate;
} * const reg_lcd = (struct reg_lcd *)0x0D000000;

volatile struct reg_misc {
	uint32_t iopCtrl;
	uint32_t iopStop;
	uint32_t iopStart;
	uint32_t iopData; // ?
	uint32_t dacCtrl;
	int32_t dacData; // s16
	uint32_t batCtrl;
} * const reg_misc = (struct reg_misc *)0x0D800000;

struct reg_timer {
	uint32_t max;
	uint32_t current;
	uint32_t flags;
};

volatile uint32_t * const reg_vram = (uint32_t *)0x0D000100;

struct __attribute__((packed, aligned(4))) alarm_info {
	uint8_t minuteBCD;
	uint8_t hourBCD;

	uint8_t enabled : 1;
	uint8_t keylock : 1;
	uint8_t attenate : 2; // 0:loud 1:med 2:mute
	uint8_t : 3; // unknown
	uint8_t goodrtc : 1; // datetime has set

	uint8_t : 8; // unknown
	uint32_t offfont; // when use, bit-or 0x04000000(KROM).
};

const static uint16_t timer_intflags[] = { IRQ_TIMER0, IRQ_TIMER1, IRQ_TIMER2 };

// @param clock 1..7, really 7 is max? (Dr.Hell and NoCash says max is 8 but 3 bit width...)
//        GUI seems use 1..7, not 0..7 nor 1..8.
INLINE void sysSetCpuClock(int clock) {
	__asm("mov r0, %0; svc 0x04" : : "rI"(clock) : "r0");
}
// return value will differ between calls, specify volatile to __asm__.
INLINE uint32_t sysGetBcdDate() { uint32_t result; __asm__ volatile("svc 0x0d; mov %0, r0" : "=r"(result) : : "r0"); return result; }
// return value will differ between calls, specify volatile to __asm__.
INLINE uint32_t sysGetBcdTime() { uint32_t result; __asm__ volatile("svc 0x0e; mov %0, r0" : "=r"(result) : : "r0"); return result; }
// note: return value of GetPtrToAlarmSetting is run-time constant. adding volatile to __asm__ makes code redundant.
INLINE struct alarm_info *sysGetPtrToAlarmSetting() { void *result; __asm__("svc 0x13; mov %0, r0" : "=r"(result) : : "r0"); return result; }

// @param timer 0..2
// @param div 0..2 (0: div2, 1: div32, 2: div512)
// @param value 0..65535, should be uint16_t...
// @remarks timer0..1 is IRQ but timer2 is FIQ.
INLINE void setTimerRaw(int timer, int div, int value) {
	volatile struct reg_timer * const reg_timern = (struct reg_timer *)(0x0A800000 + (timer << 4));
	reg_timern->flags = 0x00000000; // timerN disable div2
	reg_timern->max = value; // timerN
	reg_timern->flags = 0x00000004 | div; // timerN enable divN

	reg_intr->enable = timer_intflags[timer]; // irq enable timerN
}

// @param timer 0..2
INLINE void unsetTimer(int timer) {
	volatile struct reg_timer * const reg_timern = (struct reg_timer *)(0x0A800000 + (timer << 4));
	reg_timern->flags = 0x00000000; // timerN disable div2
	reg_intr->disable = timer_intflags[timer]; // irq disable timerN (but no effect...)
}

INLINE void setHandler(int which, void (*fptr)(void)) {
	//__asm("mov r0, %0; mov r1, %1; svc 0x01" : : "r"(which), "r"(fptr) : "r0", "r1");
	low_handler_ptrs[which] = fptr;
}

///

void stop_sound() {
	g_adpcm.psound = 0;
	reg_intr->disable = IRQ_TIMER1;
	reg_misc->iopStop = IOPPIN_SOUND;
	reg_misc->dacCtrl = DACCTRL_DISABLE;
	unsetTimer(1);
	// TODO re-set cpu clock?
}

#define start_sound(srate, start, end) do_start_sound(3997696 / 2 / (srate), start, end)
void do_start_sound(uint32_t sintervaldiv2, const uint8_t *start, const uint8_t *end) {
	if(2 <= g_attenate) {
		// mute: do nothing.
		return;
	}

	stop_sound();

	g_adpcm.psoundend = end;
	g_adpcm.psound = start;
	g_adpcm.buf = 0;
	g_adpcm.predict = 16;
	g_adpcm.presample = 0;
	sysSetCpuClock(CPUCLOCK_4MHZ);
	setTimerRaw(1, TIMER_DIV2, sintervaldiv2);
	reg_misc->iopStart = IOPPIN_SOUND;
	reg_misc->dacCtrl = DACCTRL_ENABLE;
	reg_intr->enable = IRQ_TIMER1;
}

int is_sound_end() {
	return g_adpcm.buf == 0x00 && g_adpcm.psoundend <= g_adpcm.psound;
}

int16_t decode_sound() {
	static const int16_t scale[15] = {549, 453, 375, 310, 233, 233, 233, 233, 233, 233 ,233, 310, 375, 453, 549};
	if(g_adpcm.buf == 0x00) {
		g_adpcm.buf = *g_adpcm.psound++;
	}
	uint8_t adpcm_rawsample = g_adpcm.buf & 0x0F; // u4 to u8
	g_adpcm.buf >>= 4;
	int8_t adpcm_sample = (int8_t)adpcm_rawsample - 8; // u8 to s8
	int32_t sample = g_adpcm.presample + g_adpcm.predict * adpcm_sample;
	g_adpcm.presample = sample;
	g_adpcm.predict = (int32_t)g_adpcm.predict * scale[adpcm_rawsample - 1] / 256;
	return sample;
}

///

void irq_handler(void) {
	uint32_t latch = reg_intr->latch;
	if(latch & IRQ_TIMER0) {
		g_latch = 0x80000000 | (latch & 0x3FFF);
		reg_intr->ack = IRQ_TIMER0;
	}
	if(latch & IRQ_TIMER1) { // irq latch TIMER1
		if(g_adpcm.psound == 0 || is_sound_end()) {
			stop_sound();
		} else {
			reg_misc->dacData = decode_sound() >> g_attenate;
		}
		reg_intr->ack = IRQ_TIMER1;
	}
}

///

void srand(uint32_t seed) {
	g_rand_val = 2463534242u ^ seed;
}

uint32_t rand() {
	g_rand_val = g_rand_val ^ (g_rand_val << 13) ^ (g_rand_val >> 17) ^ (g_rand_val << 5);
	return g_rand_val;
}

uint32_t rand16_up_to(uint32_t max) {
	return (rand() & 0xFFFF) * max / 0xFFFF;
}

///

#define render_image_begin \
	uint32_t srcy, srcyend, dsty; \
	/*                0           32
	 *                |           |          (srcy...srcyend -> dsty)
	 * pat.1:             <--->            : 0...image->height -> ry
	 * pat.2:     <----------->            : -ry...image->height -> 0
	 * pat.3:             <----------->    : 0...(32 - ry) -> ry
	 * pat.4:     <------------------->    : -ry...(-ry + 32) -> 0 => -ry...(32 - ry) -> 0
	 * pat.5: <--->                        : ?...? -> ?? (noop)
	 * pat.6:                         <--->: ?...? -> ?? (noop)
	 */ \
	srcy = max(0, -ry); \
	srcyend = min(max(0, 32 - ry), (int32_t)height); \
	dsty = max(0, ry); \
	while(srcy < srcyend) {

#define render_image_end \
		srcy++; \
		dsty++; \
	}

#if 0
#define render_image_opaque(name, rx, ry) render_image_opaque_(name, rx, ry)
#define render_image_opaque_(name, rx, ry) do_render_image_common(rx, ry, name##_height, name##_bit, 0)
#define render_image_transp(name, rx, ry) render_image_transp_(name, rx, ry)
#define render_image_transp_(name, rx, ry) do_render_image_common(rx, ry, name##_height, name##_bit, name##_rmask)
void do_render_image_common(int rx, int ry, uint32_t height, const uint32_t *bit, const uint32_t *rmask) {
	render_image_begin {
		if(rmask) {
			uint32_t linemask = rmask[srcy];
			uint32_t linebit = bit[srcy];
			if(rx < 0) {
				linemask = ~(linemask >> -rx);
				linebit >>= -rx;
			} else {
				linemask = ~(linemask << rx);
				linebit <<= rx;
			}
			g_offscreen[dsty] &= linemask;
			g_offscreen[dsty] |= linebit;
		} else {
			uint32_t linebit = bit[srcy];
			if(rx < 0) {
				linebit >>= -rx;
			} else {
				linebit <<= rx;
			}
			g_offscreen[dsty] = linebit;
		}
	} render_image_end
}
#else
#define render_image_opaque(name, rx, ry) render_image_opaque_(name, rx, ry)
#define render_image_opaque_(name, rx, ry) do_render_image_opaque(rx, ry, name##_height, name##_bit)
void do_render_image_opaque(int rx, int ry, uint32_t height, const uint32_t *bit) {
	render_image_begin {
		uint32_t linebit = bit[srcy];
		if(rx < 0) {
			linebit >>= -rx;
		} else {
			linebit <<= rx;
		}
		g_offscreen[dsty] = linebit;
	} render_image_end
}
#define render_image_transp(name, rx, ry) render_image_transp_(name, rx, ry)
#define render_image_transp_(name, rx, ry) do_render_image_transp(rx, ry, name##_height, name##_bit, name##_rmask)
void do_render_image_transp(int rx, int ry, uint32_t height, const uint32_t *bit, const uint32_t *rmask) {
	render_image_begin {
		uint32_t linemask = rmask[srcy];
		uint32_t linebit = bit[srcy];
		if(rx < 0) {
			linemask = ~(linemask >> -rx);
			linebit >>= -rx;
		} else {
			linemask = ~(linemask << rx);
			linebit <<= rx;
		}
		g_offscreen[dsty] &= linemask;
		g_offscreen[dsty] |= linebit;
	} render_image_end
}
#endif

void draw_a_num(int n, int x, int y) {
	uint32_t glyph = numfont[n];
	for(int i = 0; i < 5; i++) {
		g_offscreen[y] |= (glyph & 0xF) << x;
		glyph >>= 4;
		y++;
	}
}

void draw_number(uint8_t num, int x, int y) {
	uint8_t ten = num / 10;
	uint8_t one = num - ten * 10;

	if(100 <= num) {
		// special case!
		draw_a_num(1, x - 4, y);
	}
	if(0 < ten) {
		if(10 <= ten) {
			// special case!!
			ten -= 10;
		}
		draw_a_num(ten, x, y);
	}
	draw_a_num(one, x + 4, y);
}

///

void init_cards() {
	// initialize
	int i = 0;
	for(int n = 0; n < 15; i++, n++) {
		g_cards[i] = CARD_BOZU;
	}
	for(int n = 0; n < 21; i++, n++) {
		g_cards[i] = CARD_HIME;
	}
	for(int n = 0; n < 3; i++, n++) {
		g_cards[i] = CARD_ARABIC;
	}
	for(; i < 100; i++) {
		g_cards[i] = CARD_HAT;
	}

	// shuffle
	for(int i = 0; i < 99; i++) {
		int j = i + 1 + rand16_up_to(99 - i);
		uint8_t tmp = g_cards[i];
		g_cards[i] = g_cards[j];
		g_cards[j] = tmp;
	}

	g_lcards = 50;
	g_rcards = 50;

	g_mycards = 0;
	g_cpucards = 0;
	g_potcards = 0;
}

void draw_deck_cards() {
	for(int i = 0; i < (g_lcards + 15) / 16; i++) {
		render_image_transp(image_cleft, 0, 16 - i);
	}
	for(int i = 0; i < (g_rcards + 15) / 16; i++) {
		render_image_transp(image_cright, 0, 16 - i);
	}
}

void draw_card_nums() {
	draw_number(g_cpucards, 25, 0);
	draw_number(g_mycards, 25, 27);
}

void draw_pot_cards() {
	if(0 < g_potcards) {
		draw_number(g_potcards, 22 + (g_potcards < 10 ? 0 : 2), 11 - (g_potcards + 15) / 16);
		for(int i = 0; i < (g_potcards + 15) / 16; i++) {
			render_image_transp(image_cright, 0 + 8, 16 - i);
		}
	}
}

void draw_cards() {
	draw_deck_cards();
	draw_card_nums();
	draw_pot_cards();
}

///

struct {
	int y, sy;
} gs_showcardeffect;

struct {
	uint8_t thecard;
} gs_showcard;

struct {
	uint32_t sel;
} gs_pick;

void *scene_game_init(uint32_t raw, uint32_t pre);

void *scene_gameover(uint32_t raw, uint32_t pre) {
	uint32_t pressed = (pre ^ raw) & raw;

	if(g_frame < 45) {
		draw_number(g_cpucards, 25, 0);
		draw_number(g_mycards, 25, 27);
		if(g_frame < 30 || g_frame % 2 == 0) {
			draw_pot_cards();
		}
	} else if(g_frame - 45 < 13 * 4) {
		draw_number(g_cpucards, 25 - (g_frame - 45) / 4, 0 + (g_frame - 45) / 16);
		draw_number(g_mycards, 25 - (g_frame - 45) / 4, 27 - (g_frame - 45) / 16);
		if(g_frame - 45 < 13 * 4 / 2 && g_frame % 2 == 0) {
			draw_pot_cards();
		}
	} else if(g_frame - 45 - 13 * 4 < 30) {
		draw_number(g_cpucards, 12, 3);
		draw_number(g_mycards, 12, 27-3);
	} else {
		if(g_frame - 45 - 13 * 4 - 30 == 0) {
			start_sound(4000, panchi4000_adpcm_start, panchi4000_adpcm_end);
		}
		if(g_cpucards < g_mycards) {
			render_image_opaque(image_win, 0, 10);
		} else if(g_cpucards == g_mycards) {
			render_image_opaque(image_draw, 0, 9);
		} else {
			render_image_opaque(image_lose, 0, 10);
		}
		draw_number(g_cpucards, 12, 3);
		draw_number(g_mycards, 12, 27-3);
		if(pressed & IRQ_ACTION) {
			return scene_game_init;
		}
	}

	g_frame++;

	return 0;
}

void *scene_gameover_init(uint32_t raw, uint32_t pre) {
	g_frame = 0;
	return scene_gameover;
}

void *scene_pick_init(uint32_t raw, uint32_t pre);

void *scene_docard(uint32_t raw, uint32_t pre) {
	uint8_t *sidecards;

	if(g_gameside == 0) {
		sidecards = &g_mycards;
	} else {
		sidecards = &g_cpucards;
	}

	switch(gs_showcard.thecard) {
	case CARD_HIME:
		(*sidecards)++;
		(*sidecards) += g_potcards;
		g_potcards = 0;
		break;
	case CARD_BOZU:
		g_potcards += *sidecards + 1;
		*sidecards = 0;
		break;
	case CARD_ARABIC: /*FALLTHROUGH*/
	case CARD_HAT:
		(*sidecards)++;
		break;
	}
	return scene_pick_init;
}

void *scene_showcardeffect(uint32_t raw, uint32_t pre) {
	uint32_t pressed = (pre ^ raw) & raw;
	switch(gs_showcard.thecard) {
	case CARD_HIME:
		render_image_opaque(image_ctext, 0, 0);
		render_image_opaque(image_chime, 0, 15);
		render_image_transp(image_hime, 0, 6 + gs_showcardeffect.y);
		if(g_frame < 15) {
			gs_showcardeffect.y += gs_showcardeffect.sy;
			if(g_frame % 2 == 0) {
				gs_showcardeffect.sy--;
			}
		}
		break;
	case CARD_BOZU:
		render_image_opaque(image_ctext, 0, 0);
		render_image_opaque(image_cbozu, 0, 15);
		render_image_transp(image_bo, max(0, 32 - (int32_t)(g_frame * 5)) + rand16_up_to(5) - 2, 8 + rand16_up_to(5) - 2);
		render_image_transp(image_zu, max(0, 96 - (int32_t)(g_frame * 5)) + rand16_up_to(5) - 2, 7 + rand16_up_to(5) - 2);
		break;
	default:
		for(;;){/**/}
	}

	g_frame++;

	if(pressed & IRQ_ACTION) {
		stop_sound();
		return scene_docard;
	}

	return 0;
}

void *scene_showcardeffect_init(uint32_t raw, uint32_t pre) {
	switch(gs_showcard.thecard) {
	case CARD_ARABIC: /*FALLTHROUGH*/
	case CARD_HAT:
		return scene_docard;
	case CARD_HIME:
		gs_showcardeffect.y = -20;
		gs_showcardeffect.sy = 5;
		start_sound(6000, fan6000_adpcm_start, fan6000_adpcm_end);
		break;
	case CARD_BOZU:
		start_sound(6000, deden6000_adpcm_start, deden6000_adpcm_end);
		break;
	}
	g_frame = 0;
	return scene_showcardeffect;
}

void *scene_showcard(uint32_t raw, uint32_t pre) {
	int y = -32 + min(16, g_frame) * 2;
	render_image_opaque(image_ctext, 0, y);
	switch(gs_showcard.thecard) {
	case CARD_HIME:   render_image_opaque(image_chime,  0, 15 + y); break;
	case CARD_BOZU:   render_image_opaque(image_cbozu,  0, 15 + y); break;
	case CARD_ARABIC: render_image_opaque(image_chat,   0, 15 + y); break;
	case CARD_HAT:    render_image_opaque(image_ctenno, 0, 15 + y); break;
	default: for(;;){/**/}
	}

	g_frame++;
	if(64 < g_frame) {
		return scene_showcardeffect_init;
	}

	return 0;
}

void *scene_showcard_init(uint32_t raw, uint32_t pre) {
	g_frame = 0;
	start_sound(4000, ponr4000_adpcm_start, ponr4000_adpcm_end);
	return scene_showcard;
}

void *scene_showingcard(uint32_t raw, uint32_t pre) {
	uint32_t pressed = (pre ^ raw) & raw;
	int y = -48 + g_frame / 8;
	// no drawing text is required.
	switch(gs_showcard.thecard) {
	case CARD_HIME:   render_image_opaque(image_chime,  0, 15 + y); break;
	case CARD_BOZU:   render_image_opaque(image_cbozu,  0, 15 + y); break;
	case CARD_ARABIC: render_image_opaque(image_chat,   0, 15 + y); break;
	case CARD_HAT:    render_image_opaque(image_ctenno, 0, 15 + y); break;
	default: for(;;){/**/}
	}

	g_frame++;
	if(128 < g_frame || (pressed & IRQ_ACTION)) {
		return scene_showcard_init;
	}

	return 0;
}

void *scene_showingcard_init(uint32_t raw, uint32_t pre) {
	g_frame = 0;
	start_sound(4000, iyo4000_adpcm_start, iyo4000_adpcm_end);
	return scene_showingcard;
}

void pickcard(int side) {
	uint8_t *pnumcards = (side == 0) ? &g_lcards : &g_rcards;
	(*pnumcards)--;
	gs_showcard.thecard = g_cards[50 * side + *pnumcards];
	g_cards[50 * side + *pnumcards] = CARD_NONE;
}

void *scene_cpupickanim(uint32_t raw, uint32_t pre) {
	uint32_t pressed = (pre ^ raw) & raw;

	if(g_frame == 10 * 2) {
		pickcard(g_cpuside);
	}

	draw_cards();

	int thick = (((g_cpuside == 0) ? g_lcards : g_rcards) + 15) / 16 - 1;
	int x = (g_cpuside == 0) ? 0 : 9;
	if(g_frame < 10 * 2) {
		render_image_transp(image_cpufree, x, -10 + g_frame / 2 - thick);
	} else {
		render_image_transp(image_cpucatch, x, -max(0, (int32_t)(g_frame - (10 + 2) * 2)) / 2 - thick);
	}

	if(pressed & IRQ_ACTION) {
		if(g_frame < 10 * 2) {
			pickcard(g_cpuside);
		}
		return scene_showcard_init;
	}

	g_frame++;
	if((10 + (10 + 2)) * 2 < g_frame) {
		return scene_showingcard_init;
	}

	return 0;
}

void *scene_cpupickanim_init(uint32_t raw, uint32_t pre) {
	start_sound(4000, mecha4000_adpcm_start, mecha4000_adpcm_end);

	g_frame = 0;

	return scene_cpupickanim;
}

void *scene_cpumove(uint32_t raw, uint32_t pre) {
	uint32_t pressed = (pre ^ raw) & raw;

	int x;

	if(g_cpuside == 0) {
		// r->l
		x = 9 - g_frame / 4;
	} else {
		// l->r
		x = g_frame / 4;
	}

	int thick = (((g_cpuside == 0) ? g_lcards : g_rcards) + 15) / 16 - 1;
	// image_cpufree is transp image but draw as opaque: faster.
	render_image_opaque(image_cpufree, x, -10 - thick);

	// draw cards/nums after rendering opaque cpufree to avoid disappearing(overwrote) nums.
	draw_cards();

	if(pressed & IRQ_ACTION) {
		pickcard(g_cpuside);
		return scene_showcard_init;
	}

	g_frame++;
	if(9 * 4 < g_frame) {
		return scene_cpupickanim_init;
	}

	return 0;
}

void *scene_cpumove_init(uint32_t raw, uint32_t pre) {
	uint8_t newside = rand16_up_to(4) >> 1;

	if(g_lcards == 0) {
		newside = 1;
	}
	if(g_rcards == 0) {
		newside = 0;
	}

	if(newside == g_cpuside) {
		return scene_cpupickanim_init;
	}

	start_sound(4000, mecha4000_adpcm_start, mecha4000_adpcm_end);
	g_cpuside = newside;
	g_frame = 0;

	return scene_cpumove;
}

void *scene_pickanim(uint32_t raw, uint32_t pre) {
	switch(g_frame / 8) {
	case 0: draw_cards(); render_image_transp(image_hand1, gs_pick.sel * 9, 12); break;
	case 1: draw_cards(); render_image_transp(image_hand2, gs_pick.sel * 9, 11); break;
	case 2: draw_cards(); render_image_transp(image_hand3, gs_pick.sel * 9, 5); break;
	case 3: draw_cards(); render_image_transp(image_hand4, gs_pick.sel * 9, 0); break;
	case 4: render_image_opaque(image_hand5, 0, 0); break;
	case 5: render_image_opaque(image_hand6, 0, 0); break;
	default: return scene_showingcard_init;
	}
	g_frame++;
	return 0;
}

void *scene_pickanim_init(uint32_t raw, uint32_t pre) {
	pickcard(gs_pick.sel);

	g_frame = 0;
	return scene_pickanim;
}

void *scene_pick(uint32_t raw, uint32_t pre) {
	uint32_t pressed = (pre ^ raw) & raw;

	render_image_opaque(image_carrow, gs_pick.sel == 0 ? 8 : 17, 7);

	// draw carrowthen draw cards to avoid disappearing(overwrote) pot card num.
	draw_cards();

	if(pressed & IRQ_LEFT && 0 < g_lcards) {
		gs_pick.sel = 0;
	}
	if(pressed & IRQ_RIGHT && 0 < g_rcards) {
		gs_pick.sel = 1;
	}
	if(pressed & IRQ_ACTION) {
		start_sound(4000, foshu4000_adpcm_start, foshu4000_adpcm_end);
		return scene_pickanim_init;
	}

	return 0;
}

void *scene_pick_init(uint32_t raw, uint32_t pre) {
	if(g_lcards == 0 && g_rcards == 0) {
		return scene_gameover_init;
	}
	if(gs_pick.sel == 0 && g_lcards == 0) {
		gs_pick.sel = 1;
	}
	if(gs_pick.sel == 1 && g_rcards == 0) {
		gs_pick.sel = 0;
	}

	g_gameside = (g_gameside + 1) % 2;
	if(g_gameside == 0) {
		return scene_pick;
	} else {
		return scene_cpumove_init;
	}
}

void *scene_game_init(uint32_t raw, uint32_t pre) {
	init_cards();
	g_gameside = 1; // flipped first
	g_cpuside = 0;
	gs_pick.sel = 0;
	return scene_pick_init;
}

///

void appmain(void) {
	// irq disable all except COMM,IOP,RTC (used by kernel) and unknown 0x20
	// never mask RTC or IOP or COMM by app! (they are handled by kernel and it is necessary)
	// COMM mask is controlled by sysSetComOnOff.
	reg_intr->disable = 0x00003FDF & ~(IRQ_COMM | IRQ_IOP | IRQ_RTC);

	for(int i = 0; i < 32; i++) { reg_vram[i] = 0; }

	reg_lcd->ctrl = (reg_lcd->ctrl & LCDCTRL_ROTATE) |
		LCDCTRL_DMALL | LCDCTRL_ENABLE | LCDCTRL_RR64HZ | LCDCTRL_ACTIVE;

	g_attenate = sysGetPtrToAlarmSetting()->attenate;

	setHandler(HANDLER_IRQ, irq_handler); // register irq handler

	sysSetCpuClock(CPUCLOCK_4MHZ);
	//setTimerRaw(0, TIMER_DIV32, 62500); // 2Hz
	setTimerRaw(0, TIMER_DIV2, 3997696 / 2 / 64);
	unsetTimer(1);
	unsetTimer(2);

	srand(sysGetBcdDate() ^ sysGetBcdTime());

	uint32_t pre = 0;
	uint32_t action_pressed = 0;

	void *(*scene)(uint32_t, uint32_t) = scene_game_init;

	for(;;) {
		g_latch = 0;
		while(!g_latch) { /*nothing*/ }

		uint32_t raw = reg_intr->raw;

		for(int i = 0; i < 32; i++) { g_offscreen[i] = 0; }

		for(;;) {
			void *next = scene(raw, pre);
			pre = raw;
			if(!next) {
				break;
			}
			scene = next;
		}

		// appexit test
		if(raw & IRQ_ACTION) {
			if(90 < ++action_pressed) {
				break;
			}
			if(90 - 64 < action_pressed) {
				int i;

				for(i = 0; i < 16 - (90 - action_pressed) / 4 - 1; i++) {
					g_offscreen[i] = 0;
				}
				g_offscreen[i] = 0xFFFFffff;

				i = 16 + (90 - action_pressed) / 4;
				g_offscreen[i] = 0xFFFFffff;
				for(i++; i < 32; i++) {
					g_offscreen[i] = 0;
				}
			}
		} else {
			action_pressed = 0;
		}

		for(int i = 0; i < 32; i++) { reg_vram[i] = g_offscreen[i]; }
	}

	unsetTimer(0);
	stop_sound();

	for(int i = 0; i < 32; i++) { reg_vram[i] = 0; }
	while(reg_intr->raw & IRQ_ACTION) { /*nothing*/ }
}
